(*******************************************************************************

  Cryptographically Faithful Verification of Security Protocols

  Module:  VerifTools/EvaluationLogic.thy (Isabelle/HOL 2005)
  ID:      $Id: EvaluationLogic.thy 29266 2006-08-10 15:07:15Z csprenge $
  
  A dynamic logic for monadic programs

  Copyright (C) 2004-2006 Christoph Sprenger <christoph.sprenger@inf.ethz.ch>
  Licence: LGPL

*******************************************************************************)


header {* Evaluation Logic *} 

theory EvaluationLogicR imports RMonad RMonad_Decomposition begin

text {* Andy Pitts' evaluation logic is a dynamic logic for monadic programs. *}


(******************************************************************************)
subsection {* Side effect-free programs *}
(******************************************************************************)

text {* The definition and the first two lemmas are monad-independent. *}

lemma seqN_alter: "p \<guillemotright>N q = (\<lambda>s. \<Union> (q o snd) ` p s)"
  by(auto simp add: bindN_def bindS_def fun_eq_iff)

lemma bindN_empty:
  "p s = {} \<Longrightarrow> (p \<guillemotright>=N q) s = {}"
  by(auto simp add: bindN_def bindS_def)

lemma seqN_empty:
  "p s = {} \<Longrightarrow> (p \<guillemotright>N q) s = {}"
  by(auto simp add: bindN_empty)

definition
  sefN :: "('a, 's) nds_monad \<Rightarrow> bool" where
  "sefN p \<equiv> (p \<guillemotright>N returnN ()) = returnN ()"

lemma sefN_return [simp]: "sefN (returnN x)"
  by (simp add: bindN_left_unit sefN_def)

lemma sefN_bind [simp]: 
  assumes A: "sefN p" 
  shows "(p \<guillemotright>N q) = q" using A
proof -
  have "(p \<guillemotright>N q) = (p \<guillemotright>N returnN ()) \<guillemotright>N q" by(simp add: bindN_left_unit bindN_assoc)
  also from A have "... = (returnN () \<guillemotright>N q)" by (simp add: sefN_def)
  also have "... = q" by (simp add: bindN_left_unit)
  finally show ?thesis .
qed

lemma sefN_not_empty:
  "sefN p \<Longrightarrow> p s \<noteq> {}"
proof 
  assume sef: "sefN p" and empty: "p s = {}"
  thm seqN_empty[where p=p and s=s, OF empty]
  have "(p \<guillemotright>N returnN ()) s = {}" using seqN_empty[where p=p and s=s, OF empty]
    by(force)
  thus False using sef
    by(auto simp add: sefN_def returnN_def returnS_def)
qed

lemma sefN_not_two:
  "sefN p \<Longrightarrow> \<forall>s'. s' \<in> snd ` p s \<longrightarrow> s' = s"
proof-
  assume sef: "sefN p"
  {
    fix s'
    assume mem: "s' \<in> snd ` p s"
    hence in_res: "s' \<in> snd ` (p \<guillemotright>N returnN ()) s" 
      by(force simp add: seqN_alter returnN_def returnS_def)
    hence "s' = s"
      using sefN_bind[THEN fun_cong, where x=s and g="returnN ()", OF sef]
      by(simp add: returnN_def returnS_def)
  } 
  thus ?thesis
    by(auto)
qed

lemma singletonI2: "\<lbrakk> A \<noteq> {}; \<forall>x1. x1 \<in> A \<longrightarrow> x1 = x \<rbrakk> \<Longrightarrow> A = {x}"
  by(blast)

lemma sefN_singleton:
  assumes A: "sefN p"
  shows "snd ` p s = {s}"
  using sefN_not_empty[OF A] sefN_not_two[OF A]
  by(auto intro!: singletonI2)

text {* A characterisation of side-effect freeness for our particular monad. *}

lemma sefN_concrete: "sefN p = (\<forall> s. \<exists> A. A \<noteq> {} \<and> p s = A \<times> {s})"
proof
  assume sef: "sefN p"
  show "\<forall>s. \<exists>A. A \<noteq> {} \<and> p s = A \<times> {s}"
  proof
    fix s
    show "\<exists>A. A \<noteq> {} \<and> p s = A \<times> {s}" using sefN_singleton[OF sef, where s=s]
      apply(rule_tac x="fst ` p s" in exI)
      apply(force)
      done
  qed
next
  assume single: "\<forall>s. \<exists>A. A \<noteq> {} \<and> p s = A \<times> {s}"
  {
    fix s
    obtain A where A_def: "A \<noteq> {} \<and> p s = A \<times> {s}" using single by force
    from this obtain x where x_in_A: "x \<in> A" by auto
    hence "(\<Union>x\<in>p s. {((), snd x)}) = {((), s)}" using A_def
      by(intro Equiv_Relations.UN_constant_eq, auto)
  }
  thus "sefN p"
    by(auto simp add: sefN_def seqN_alter returnN_def returnS_def)
qed

abbreviation
  "seqR" :: "('a, 's) r_monad \<Rightarrow> ('b, 's) r_monad \<Rightarrow> ('b, 's) r_monad"    (infixl "\<guillemotright>R" 60) where
  "m1 \<guillemotright>R m2 \<equiv> m1 \<guillemotright>=R (\<lambda>_. m2)" 

lemma bindR_empty:
  "p s = {} \<Longrightarrow> (p \<guillemotright>=R q) s = {}"
  by(auto simp add: bindR_def bindN_empty)

lemma seqR_empty:
  "p s = {} \<Longrightarrow> (p \<guillemotright>R q) s = {}"
  by(auto simp add: bindR_empty)

definition
  sefR :: "('a, 's) r_monad \<Rightarrow> bool" where
  "sefR p \<equiv> (p \<guillemotright>R returnR ()) = returnR ()"

lemma sefR_return [simp]: "sefR (returnR x)"
  by(simp add: bindR_left_unit sefR_def)

lemma sefR_bind [simp]: 
  assumes A: "sefR p" 
  shows "(p \<guillemotright>R q) = q" using A
proof -
  have "(p \<guillemotright>R q) = (p \<guillemotright>R returnR ()) \<guillemotright>R q" by(simp add: bindR_left_unit bindR_assoc)
  also from A have "... = (returnR () \<guillemotright>R q)" by (simp add: sefR_def)
  also have "... = q" by (simp add: bindR_left_unit)
  finally show ?thesis .
qed

lemma some_notin_restart[simp]: "(Some r, s') \<notin> restart s"
  by(auto simp add: restart_def returnN_def returnS_def)

lemma none_notin_returnR[simp]: "(None, s') \<notin> returnR x s"
  by(auto simp add: returnR_def returnN_def returnS_def)

lemma option_in: "\<lbrakk> x \<notin> A a; x \<in> (option_case A B opt) a \<rbrakk> \<Longrightarrow> 
  \<exists>z. x \<in> B z a"
  by(case_tac opt, auto)

lemma bindR_bindN: "None \<notin> fst ` p s \<Longrightarrow> (p \<guillemotright>=R q) s = (p \<guillemotright>=N (q o the)) s"
  apply(auto simp add: bindR_def liftR_def bindN_def bindS_def restart_def)
  apply(case_tac[!] aa, force+)
  done

lemma bindR_bindN': "\<forall>s. None \<notin> fst ` p s \<Longrightarrow> (p \<guillemotright>=R q) = (p \<guillemotright>=N (q o the))"
  by(rule ext, simp add: bindR_bindN)

lemma bindR_bindN'': "(\<And>s. None \<notin> fst ` p s) \<Longrightarrow> (p \<guillemotright>=R q) = (p \<guillemotright>=N (q o the))"
  by(simp add: bindR_bindN')

lemma seqR_seqN: "(None \<notin> fst ` p s) \<Longrightarrow> (p \<guillemotright>R q) s = (p \<guillemotright>N q) s"
  apply(auto simp add: bindR_def liftR_def bindN_def bindS_def restart_def)
  apply(case_tac[!] aa, force+)
  done

lemma seqR_seqN': "(\<And>s. None \<notin> fst ` p s) \<Longrightarrow> (p \<guillemotright>R q) = (p \<guillemotright>N q)"
  by(rule ext, simp add: seqR_seqN)

lemma sefR_None:
  "sefR p \<Longrightarrow> None \<notin> fst ` p s"
proof-
  assume "sefR p"
  hence "None \<notin> fst ` (p \<guillemotright>R returnR ()) s"
    by(auto simp add: sefR_def)
  thus ?thesis using bindR_none_decompose[where p=p and q="\<lambda>_. returnR()" and s=s]
    by(force)
qed

lemma sefR_not_empty:
  "sefR p \<Longrightarrow> p s \<noteq> {}"
proof 
  assume sef: "sefR p" and empty: "p s = {}"
  have "(p \<guillemotright>R returnR ()) s = {}" using seqR_empty[where p=p and s=s, OF empty]
    by(auto)
  thus False using sef
    by(auto simp add: sefR_def returnR_def returnN_def returnS_def)
qed

lemma sefR_not_two:
  "sefR p \<Longrightarrow> \<forall>s'. s' \<in> snd ` p s \<longrightarrow> s' = s"
proof-
  assume sef: "sefR p"
  {
    fix s'
    assume mem: "s' \<in> snd ` p s"
    from this obtain x where x_in_p: "(Some x, s') \<in> p s" using sefR_None[OF sef, where s=s]
      apply(auto)
      apply(case_tac a)
      apply(force)+
      done
    hence in_res: "(Some (), s') \<in> (p \<guillemotright>R returnR ()) s" 
      using sef[THEN sefR_None, THEN seqR_seqN'[where p=p and q="returnR ()"]]
      apply(auto simp add: seqN_alter returnR_def dest!: fun_eq_iff[THEN iffD1])
      apply(rule_tac x="(Some x, s')" in bexI)
      apply(simp_all add: returnN_def returnS_def)
      done
    hence "s' = s"
      using sefR_bind[THEN fun_cong, where x=s and g="returnR ()", OF sef]
      by(simp add: returnR_def returnN_def returnS_def)
  } 
  thus ?thesis
    by(auto)
qed

lemma sefR_singleton:
  assumes A: "sefR p"
  shows "snd ` p s = {s}"
  using sefR_not_empty[OF A] sefR_not_two[OF A]
  by(auto intro!: singletonI2)

lemma image_pointwise: "\<forall>x \<in> A. \<exists>y. x = f y \<Longrightarrow> \<exists>C. A = f ` C"
proof-
  assume "\<forall>x \<in> A. \<exists>y. x = f y"
  hence "\<exists>g. \<forall>x \<in> A. x = f (g x)"
    by(rule bchoice)
  from this obtain g where "\<forall>x \<in> A. x = f (g x)" ..
  hence "A = f ` g ` A"
    by(auto)
  thus ?thesis
    by auto
qed

lemma not_None_image_Some: "None \<notin> A \<Longrightarrow> \<exists>B. A = Some ` B"
proof-
  assume "None \<notin> A"
  hence "\<forall>x \<in> A. \<exists>y. x = Some y"
    by(auto, case_tac x, auto)
  thus ?thesis
    by(rule image_pointwise)
qed
    
lemma sefR_concrete: "sefR p = (\<forall> s. \<exists> A. A \<noteq> {} \<and> p s = (Some ` A) \<times> {s})"
proof
  assume sef: "sefR p"
  show "\<forall>s. \<exists>A. A \<noteq> {} \<and> p s = (Some ` A) \<times> {s}"
  proof
    fix s
    def A \<equiv> "fst ` p s"
    have "\<exists>B. A = Some ` B" using sefR_None[OF sef, where s=s]
      by(auto simp add: A_def intro!: not_None_image_Some)
    from this obtain B where B_def: "A = Some ` B" ..
    have req: "A \<noteq> {} \<and> p s = A \<times> {s}" using sefR_singleton[OF sef, where s=s] 
      by(force simp add: A_def)
    hence "B \<noteq> {} \<and> p s = Some ` B \<times> {s}"
      by(simp add: B_def)
    thus "\<exists>A. A \<noteq> {} \<and> p s = Some ` A \<times> {s}"  by auto
  qed
next
  assume single: "\<forall>s. \<exists>A. A \<noteq> {} \<and> p s = Some ` A \<times> {s}"
  {
    fix s
    have "None \<notin> fst ` p s" using single
      by(force)
  } note seqRN = seqR_seqN'[where p=p, OF this]
  {
    fix s
    obtain A where A_def: "A \<noteq> {} \<and> p s = Some ` A \<times> {s}" using single by force
    from this obtain x where x_in_A: "x \<in> A" by auto
    have "(p \<guillemotright>R returnR ()) s = (\<Union>x\<in>p s. {(Some (), snd x)})" using seqRN[where q="returnR ()"]
      by(auto simp add: seqN_alter returnR_def returnN_def returnS_def)
    moreover have "... = {(Some (), s)}" using A_def x_in_A
      by(intro Equiv_Relations.UN_constant_eq, auto)
    finally have "(p \<guillemotright>R returnR ()) s = {(Some (), s)}" by auto
  }
  thus "sefR p"
    by(auto simp add: sefR_def returnR_def returnN_def returnS_def intro!: ext)
qed

lemma sef_chooseN:
  "A \<noteq> {} \<Longrightarrow> chooseN A \<guillemotright> m = m"
  apply(rule ext)
  apply(rule set_eqI)
  apply(case_tac xa)
  apply(case_tac a)
  apply(force simp add: unfold_monad chooseN_def choose_def image_def bind_none_decompose
    split: option.split_asm sum.split_asm)
  apply(case_tac aa)
  apply(force simp add: unfold_monad chooseN_def choose_def image_def 
    bind_norm_decompose bind_exc_decompose
    split: option.split_asm sum.split_asm)+
  done

lemma add_chooseN: "A \<noteq> {} \<Longrightarrow> m \<guillemotright> m' = (m \<guillemotright> chooseN A \<guillemotright> m')"
  by(simp add: sef_chooseN)

(******************************************************************************)
subsection {* Lemmas about program composition operators *}
(******************************************************************************)

(*
text {* These are low-level lemmas breaking open the state-exception monad. *}

lemma bind_norm_decompose:
  "((p >>= q) s = (Value c, u)) =
   (\<exists> b t. (p s) = (Value b, t) & (q b t) = (Value c, u))" 
apply (auto simp add: bind_def split_def Let_def)
apply (subgoal_tac "\<exists> x y. (p s) = (x, y)", safe, auto)
done

lemma bind_exc_decompose:
  "((p >>= q) s = (Exception, t)) =
   ((\<exists> b u.  (p s) = (Value b, u) & (q b u) = (Exception, t)) | 
    (p s) = (Exception, t))"
apply (auto simp add: bind_def split_def Let_def)
apply (subgoal_tac "\<exists> x y. (p s) = (x, y)", safe, auto)+
done

lemma try_catch_norm_decompose: 
  "((try_catch p q s) = (Value x, t)) =
   (p s = (Value x, t) | (\<exists> u. (p s) = (Exception, u) & (q u) = (Value x, t)))"
apply (auto simp add: try_catch_def split_def Let_def)
apply (subgoal_tac "\<exists> x y. (p s) = (x, y)", safe, auto)+
done

lemma try_catch_exc_decompose: 
  "((try_catch p q s) = (Exception, t)) =
   (\<exists> u. (p s) = (Exception, u) & (q u) = (Exception, t))"
apply (auto simp add: try_catch_def split_def Let_def)
apply (subgoal_tac "\<exists> x y. (p s) = (x, y)", safe, auto)+
done

*)


(******************************************************************************)
subsection {* Evaluation logic *}
(******************************************************************************)

text {* Here we define a generalised dynamic logic similar to Pitts'
\textit{evaluation logic}. This will serve as a weakest precondition calculus
to (hopefullly) automate many later program correctness proofs. *}

text {* State assertions are just sets of states. Also, give the bottom and top
element of the powerset lattice more suggestive names. *}

type_synonym
  's assn = "'s set"

abbreviation (input)
  "true" :: "'a set" where
  "true \<equiv> UNIV"

abbreviation (input)
  "false" :: "'a set" where
  "false \<equiv> {}"

text {* We have two (box) modalities, which we dub \textit{normal} and
\textit{exception} modalities. We do not define the dual diamond modalities,
since we do not need them later. *}

definition Box :: "('a, 's)nds_monad \<Rightarrow> ('a \<Rightarrow> 's assn) \<Rightarrow> 's assn" where
  "Box m Q \<equiv> {s. \<forall>a s'. (a, s') \<in> m s \<longrightarrow> s' \<in> Q a}"

syntax
  "_Box" :: "pttrn \<Rightarrow> ('a', 's) nds_monad \<Rightarrow> ('a \<Rightarrow> 's assn) \<Rightarrow> 's assn"  
                                     ("(3[N _/ \<leftarrow>/ _] _)" [0, 0, 80] 80)

translations 
  "[N x \<leftarrow> m] Q" == "CONST Box m (\<lambda>x. Q)"

definition
  BoxS :: "('s \<Rightarrow> 's nd_monad) \<Rightarrow> 's assn \<Rightarrow> 's assn" ("(2[S/ _] _)" [0, 80] 80) where
  "BoxS m Q \<equiv> {s. \<forall>s'. s' \<in> m s \<longrightarrow> s' \<in> Q }"

definition
  BoxN :: "('a, 's) r_monad \<Rightarrow> ('a \<Rightarrow> 's assn) \<Rightarrow> 's assn" where
  "BoxN m Q \<equiv> {s. \<forall> x t. (Some x, t) \<in> (m s) \<longrightarrow> t \<in> Q x}"

definition
  BoxR :: "('a, 's) r_monad \<Rightarrow> 's assn \<Rightarrow> 's assn"    ("(2[@R/ _] _)" [0, 80] 80) where
  "BoxR m Q \<equiv> { s. \<forall> t. (None, t) \<in> (m s) \<longrightarrow> t \<in> Q}"

syntax
  "_BoxN" :: "pttrn \<Rightarrow> ('a', 's) r_monad \<Rightarrow> ('a \<Rightarrow> 's assn) \<Rightarrow> 's assn"  
                                     ("(3[R _/ \<leftarrow>/ _] _)" [0, 0, 80] 80)

translations 
  "[R x \<leftarrow> m] Q" == "CONST BoxN m (\<lambda>x. Q)"


definition liftPredNorm :: "('a \<Rightarrow> 's assn) \<Rightarrow> 'a option \<Rightarrow> 's assn" where
  "liftPredNorm P \<equiv> \<lambda>x. (case x of
    None \<Rightarrow> UNIV
    | Some val \<Rightarrow> P val)"

definition liftPredRes :: "('s assn) \<Rightarrow> 'a option \<Rightarrow> 's assn" where
  "liftPredRes P \<equiv> \<lambda>x. (case x of
    None \<Rightarrow> P 
    | Some _ \<Rightarrow> UNIV)"

lemma BoxN_thru_Box:
  "BoxN m P = Box m (liftPredNorm P)"
  by(auto simp add: BoxN_def Box_def liftPredNorm_def split: option.split)

lemma BoxR_thru_Box:
  "BoxR m P = Box m (liftPredRes P)"
  by(auto simp add: BoxR_def Box_def liftPredRes_def split: option.split)

definition liftPredVal :: "('a \<Rightarrow> 's assn) \<Rightarrow> 'e + 'a \<Rightarrow> 's assn" where
  "liftPredVal P \<equiv> \<lambda>x. (case x of
    Inl exc \<Rightarrow> UNIV
    | Inr val \<Rightarrow> P val)"

definition liftPredExc :: "('e \<Rightarrow> 's assn) \<Rightarrow> 'e + 'a \<Rightarrow> 's assn" where
  "liftPredExc P \<equiv> \<lambda>x. (case x of
    Inl exc \<Rightarrow> P exc
    | Inr val \<Rightarrow> UNIV)"

lemma[simp]: "liftPredVal (\<lambda>x. true) = (\<lambda>x. true)"
  by(simp add: liftPredVal_def)

lemma[simp]: "liftPredExc (\<lambda>x. true) = (\<lambda>x. true)"
  by(simp add: liftPredExc_def)

definition pred_conj :: "('a \<Rightarrow> 'b set) \<Rightarrow> ('a \<Rightarrow> 'b set) \<Rightarrow> ('a \<Rightarrow> 'b set)" where
  "pred_conj P Q \<equiv> \<lambda>x. P x \<inter> Q x"

(*
definition pred_Conj :: "'i set \<Rightarrow> ('a \<Rightarrow> 'i \<Rightarrow> 'b set) \<Rightarrow> ('a \<Rightarrow> 'i \<Rightarrow> 'b set) \<Rightarrow> ('a \<Rightarrow> 'b set)" where
*)

lemma liftVal_conj: "liftPredVal (pred_conj P Q) = pred_conj (liftPredVal P) (liftPredVal Q)"
  by(auto simp add: pred_conj_def liftPredVal_def split:sum.split intro!: ext)

lemma liftExc_conj: "liftPredExc (pred_conj P Q) = pred_conj (liftPredExc P) (liftPredExc Q)"
  by(auto simp add: pred_conj_def liftPredExc_def split:sum.split intro!: ext)

(*
lemma "liftPredVal (\<lambda>x. \<Inter> i\<in>A. P i x) = (\<lambda>x. (\<Inter> i\<in>A. liftPredVal (P i)))"
*)

definition BoxNN :: "('a, 'e, 's) re_monad \<Rightarrow> ('a \<Rightarrow> 's assn) \<Rightarrow> 's assn" where
  "BoxNN m Q \<equiv> BoxN m (liftPredVal Q)"

definition BoxNE :: "('a, 'e, 's) re_monad \<Rightarrow> ('e \<Rightarrow> 's assn) \<Rightarrow> 's assn" where
  "BoxNE m Q \<equiv> BoxN m (liftPredExc Q)"

syntax
  "_BoxNN" :: "pttrn \<Rightarrow> ('a', 'e, 's) re_monad \<Rightarrow> ('a \<Rightarrow> 's assn) \<Rightarrow> 's assn"  
                                     ("(3[ _/ \<leftarrow>/ _] _)" [0, 0, 80] 80)

translations 
  "[x \<leftarrow> m] Q" == "CONST BoxNN m (\<lambda>x. Q)"

syntax
  "_BoxNE" :: "pttrn \<Rightarrow> ('a', 'e, 's) re_monad \<Rightarrow> ('e \<Rightarrow> 's assn) => 's assn"  
                                     ("(3[@ _/ \<leftarrow>/ _] _)" [0, 0, 80] 80)

translations 
  "[@ x \<leftarrow> m] Q" == "CONST BoxNE m (\<lambda>x. Q)"

lemmas Box_defs = BoxN_def BoxR_def BoxNN_def BoxNE_def liftPredExc_def liftPredVal_def

subsubsection {* General properties of the box modalities *}

lemma Box_monotone:
  "\<forall> x. P x \<subseteq> Q x ==> [N x \<leftarrow> m] P x \<subseteq> [N x \<leftarrow> m] Q x"
  by (auto simp add: Box_def)
  
lemma BoxN_monotone: 
  "\<forall> x. P x \<subseteq> Q x ==> [R x \<leftarrow> m] P x \<subseteq> [R x \<leftarrow> m] Q x"
  by (auto simp add: BoxN_def)

lemma BoxR_monotone: "P \<subseteq> Q ==> [@R m] P \<subseteq> [@R m] Q"
  by (auto simp add: BoxR_def)

lemma sum_monoL: "\<forall>x. P x \<subseteq> Q x \<Longrightarrow> \<forall>x. (sum_case P C2 x) \<subseteq> (sum_case Q C2 x)"
  by(auto split: sum.split)

lemma sum_monoR: "\<forall>x. P x \<subseteq> Q x \<Longrightarrow> \<forall>x. (sum_case C1 P x) \<subseteq> (sum_case C1 Q x)"
  by(auto split: sum.split)

lemma BoxNN_monotone: 
  "\<forall> x. P x \<subseteq> Q x ==> [x \<leftarrow> m] P x \<subseteq> [x \<leftarrow> m] Q x"
  by(auto simp add: BoxNN_def BoxN_monotone liftPredVal_def sum_monoR)

lemma BoxNE_monotone: "\<forall>x. P x \<subseteq> Q x \<Longrightarrow> [@ x \<leftarrow> m] P x \<subseteq> [@ x \<leftarrow> m] Q x"
  by(auto simp add: BoxNE_def BoxN_monotone liftPredExc_def sum_monoL)

lemma Box_true: "[N x \<leftarrow> p] true = true"
  by (simp add: Box_def)

lemma BoxN_true: "[R x \<leftarrow> p] true = true"
  by (simp add: BoxN_def)

lemma BoxR_true: "[@R p]true = true"
  by (simp add: BoxR_def)

lemma BoxNN_true: "[ x \<leftarrow> p] true = true"
  by(simp add: BoxNN_def BoxN_true)

lemma BoxNE_true: "[@ x \<leftarrow> p]true = true"
by (simp add: BoxNE_def BoxN_true)

text {* Note that the analogue for disjunction doesn't hold for the non-deterministic monad *}

lemma BoxN_conj: 
  "[R x \<leftarrow> p]((P x) \<inter> (Q x)) = [R x \<leftarrow> p](P x) \<inter> [R x \<leftarrow> p](Q x)"
by (auto simp add: BoxN_def)

lemma BoxR_conj: "[@R p](P \<inter> Q) = [@R p]P \<inter> [@R p]Q"
by (auto simp add: BoxR_def)

lemma BoxNN_conj: 
  "[ x \<leftarrow> p]((P x) \<inter> (Q x)) = [ x \<leftarrow> p](P x) \<inter> [ x \<leftarrow> p](Q x)"
by (auto simp add: BoxNN_def liftVal_conj[simplified pred_conj_def] BoxN_conj)

lemma BoxNE_conj: 
  "[@ x \<leftarrow> p]((P x) \<inter> (Q x)) = [@ x \<leftarrow> p](P x) \<inter> [@ x \<leftarrow> p](Q x)"
by (auto simp add: BoxNE_def liftExc_conj[simplified pred_conj_def] BoxN_conj)

lemma BoxN_Conj:
  "[R x \<leftarrow> p](\<Inter> i\<in>A. P i x) = (\<Inter> i\<in>A. [R x \<leftarrow> p](P i x))"
by (auto simp add: BoxN_def)

lemma BoxR_Conj:
  "[@R p](\<Inter> i\<in>A. P i) = (\<Inter> i\<in>A. [@R p](P i))"
by (auto simp add: BoxR_def)

lemma liftVal_Conj: "liftPredVal (\<lambda>x. \<Inter>i\<in>A. P i x) = (\<lambda>x. \<Inter>i\<in>A. (liftPredVal (P i) x))"
  by(clarsimp simp add: liftPredVal_def split: sum.split intro!: ext)

lemma liftExc_Conj: "liftPredExc (\<lambda>x. \<Inter>i\<in>A. P i x) = (\<lambda>x. \<Inter>i\<in>A. (liftPredExc (P i) x))"
  by(clarsimp simp add: liftPredExc_def split: sum.split intro!: ext)

lemma BoxNN_Conj:
  "[ x \<leftarrow> p](\<Inter> i\<in>A. P x i) = (\<Inter> i\<in>A. [ x \<leftarrow> p](P x i))"
  by(simp add: BoxNN_def liftVal_Conj BoxN_Conj)

lemma BoxNE_Conj:
  "[ x \<leftarrow> p](\<Inter> i\<in>A. P x i) = (\<Inter> i\<in>A. [ x \<leftarrow> p](P x i))"
  by(simp add: BoxNN_def liftVal_Conj BoxN_Conj)

lemma BoxN_false: "[R x \<leftarrow> p] false = {s. \<forall>x. Some x \<notin> fst ` (p s) }"
  by(force simp add: BoxN_def)

lemma BoxR_false: "[@R p] false = {s. None \<notin> fst ` (p s) }"
  by(force simp add: BoxR_def)

lemma BoxNN_false: "[ x \<leftarrow> p] false = {s. \<forall>x. Some (Inr x) \<notin> fst ` (p s) }"
  by(force simp add: Box_defs BoxN_false split: sum.split)

lemma BoxNE_false: "[@ x \<leftarrow> p] false = {s. \<forall>x. Some (Inl x) \<notin> fst ` (p s) }"
  by(force simp add: Box_defs BoxN_false split: sum.split)

subsubsection {* Basic properties related to monad *}

lemma Box_return: "[N x \<leftarrow> returnN a] Q x = Q a"
  by (simp add: Box_def returnN_def returnS_def)

lemma BoxN_return: "[R x \<leftarrow> returnR a] Q x = Q a"
  by (simp add: BoxN_def returnR_def returnN_def returnS_def)

lemma BoxR_return: "[@R (returnR a)] P = true" 
  by (simp add: BoxR_def return_def)

lemma BoxNN_return: "[ x \<leftarrow> return a] Q x = Q a"
  apply(simp only: BoxNN_def BoxN_return return_def)
  apply(clarsimp simp add: BoxN_return liftPredVal_def)
  done

lemma BoxNE_return: "[@ x \<leftarrow> return a] Q x = true"
  apply(simp only: BoxNE_def BoxN_return return_def)
  apply(clarsimp simp add: BoxN_return liftPredExc_def)
  done

lemma Box_bind: "[N y \<leftarrow> p \<guillemotright>=N q] Q y = [N x \<leftarrow> p][N y \<leftarrow> q x] Q y"
  by (auto simp add: Box_def bindN_compose)

lemma BoxN_bind: "[R y \<leftarrow> p \<guillemotright>=R q] Q y = [R x \<leftarrow> p][R y \<leftarrow> q x] Q y"
  by (auto simp add: BoxN_def bindR_some_decompose)

lemma BoxR_bind: "[@R p \<guillemotright>=R q] Q = [R x \<leftarrow> p][@R q x] Q \<inter> [@R p] Q "
  by (auto simp add: BoxR_def BoxN_def bindR_none_decompose)

lemma BoxNN_bind : "[ y \<leftarrow> p \<guillemotright>= q] Q y = [ x \<leftarrow> p][ y \<leftarrow> q x] Q y"
  by(auto simp add: Box_defs bind_norm_decompose split: sum.split)

lemma BoxNE_bind : "[@ y \<leftarrow> p \<guillemotright>= q] Q y = [@ y \<leftarrow> p] Q y \<inter> [ x \<leftarrow> p][@ y \<leftarrow> q x] Q y"
  by(auto simp add: Box_defs bind_exc_decompose split: sum.split)

subsubsection {* Side-effect free programs *}

lemma BoxN_sef: "sefR m \<Longrightarrow> [R z \<leftarrow> m]P = P "
  apply (auto simp add: BoxN_def sefR_concrete)
  apply (drule_tac x=x in spec, auto)
done

lemma BoxN_sef_dep: 
  "sefR m \<Longrightarrow> [R z \<leftarrow> m](P z) = {s. s \<in> \<Inter>P ` (the ` fst ` (m s)) }"
apply (auto simp add: BoxN_def sefR_concrete)
apply (drule_tac x=x in spec, auto)+
done

lemma BoxR_sef: "sefR m ==> [@R m]P = true"
  by(auto simp add: BoxR_def sefR_concrete)

subsubsection {* Properties related to state transformers *}

(*
lemma BoxN_distill [simp]: "[s \<leftarrow> distill f](P s) = {t. t: P (f t)}"
by (simp add: BoxN_def distill_def return_def)

lemma BoxR_distill [simp]: "[@ distill f] P = true"
by (simp add: BoxR_def distill_def return_def)

lemma BoxN_assign [simp]: "[ z \<leftarrow> assign t](Q z) = {s. t: Q ()}"
by (simp add: BoxN_def assign_def return_def)

lemma BoxR_assign [simp]: "[@ assign t] P = true"
by (simp add: BoxR_def assign_def return_def)

lemma BoxN_xform [simp]: "[ z \<leftarrow> xform f](P z) = f-`(P ())"
by (simp add: xform_def, fast)

lemma BoxR_xform [simp]: "[@ xform f] P = true"
by (simp add: xform_def)
*)


subsubsection {* Properties related to exceptions *}

lemma BoxN_restart: "[R z \<leftarrow> restart ](P z) = true"
by (simp add: BoxN_def)

lemma BoxR_restart: "[@R restart]P = P"
  by (simp add: BoxR_def restart_def returnN_def returnS_def)

lemma BoxNN_restart: "[ x \<leftarrow> restart ] P x = true"
  by(auto simp add: BoxNN_def BoxN_restart)

lemma BoxNE_restart: "[@ x \<leftarrow> restart ] P x = true"
  by(auto simp add: BoxNE_def BoxN_restart)

lemma BoxNN_throw: "[ z \<leftarrow> throw e ](P z) = true"
  by(auto simp add: BoxNN_def BoxN_def throw_def returnR_def returnN_def returnS_def liftPredVal_def)

lemma BoxNE_throw: "[@ z \<leftarrow> throw e ](P z) = P e"
  by(auto simp add: BoxNE_def BoxN_def throw_def returnR_def returnN_def returnS_def liftPredExc_def)

text {* The next statement is dual to @{term "BoxR_bind"} with respect to the
mode of termination (normal or with an exception). *}

lemma BoxN_try_catch: 
  "[R x \<leftarrow> try_catchR p q](P x) = [R x \<leftarrow> p](P x) \<inter> [@R p][R x \<leftarrow> q](P x)"
  by(auto simp add: BoxR_def BoxN_def try_catchR_some_decompose)

lemma BoxR_try_catch: "[@R try_catchR p q]Q = [@R p][@R q]Q"
  by(auto simp add: BoxR_def try_catchR_none_decompose)

lemma BoxNN_try_catch: 
  "[ x \<leftarrow> try_catch p q](P x) = [ x \<leftarrow> p](P x) \<inter> [@ e \<leftarrow> p][ x \<leftarrow> q e](P x)"
  by(auto simp add: try_catch_norm_decompose Box_defs
    split: sum.split sum.split_asm)

lemma BoxNE_try_catch: 
  "[@ x \<leftarrow> try_catch p q](P x) = [@ e \<leftarrow> p][@ x \<leftarrow> q e](P x)"
  by(auto simp add: try_catch_exc_decompose Box_defs
    split: sum.split sum.split_asm)


(*
lemma BoxN_enforceb [simp]: "[x \<leftarrow> enforceb b](Q x) = {s. s: Q () | ~b}"
by (simp add: enforceb_def)

lemma BoxR_enforceb [simp]: "[@ enforceb b]Q = {s. s: Q | b}"
by (simp add: enforceb_def)

lemma BoxN_enforce [simp]: "[x \<leftarrow> enforce c](Q x) = {s. (c s) --> s : Q ()}"
by (auto simp add: enforce_def)

lemma BoxR_enforce [simp]: "[@ enforce c]Q = {s. ~(c s) --> s : Q}"
by (auto simp add: enforce_def)


lemma BoxN_noexc [simp]: 
  "[x \<leftarrow> noexc m](P x) = [ x \<leftarrow> m] P (Some x) Int [@ m] P None"
by (simp add: noexc_def)

lemma BoxR_noexc [simp]: "[@ noexc m]P = true"
by (simp add: noexc_def)
*)


subsubsection {* Linear control structures *}

text {* NOTE: unused so far; "activate" by inclusion in simp/clasets. *}

lemma BoxN_if_then_else: 
  "[R x \<leftarrow> if b then m1 else m2] (Q x) = 
   (if b then [R x \<leftarrow> m1] Q x else [R x \<leftarrow> m2] Q x)"
by (simp)

lemma BoxR_if_then_else: 
  "[@R if b then m1 else m2] Q = 
   (if b then [@R m1] Q else [@R m2] Q)"
by (simp)

lemma BoxNN_if_then_else: 
  "[ x \<leftarrow> if b then m1 else m2] (Q x) = 
   (if b then [ x \<leftarrow> m1] Q x else [ x \<leftarrow> m2] Q x)"
by (simp)

definition K_set :: "('a \<Rightarrow> bool) \<Rightarrow> 'a \<Rightarrow> 'b set" where
  "K_set P \<equiv> \<lambda>x. {s. P x}"

(*
lemma BoxNN_if_then_else: 
  "[ x \<leftarrow> if b then m1 else m2] (Q x) = 
   ([ x \<leftarrow> m1] K b \<inter> [ x \<leftarrow> m1 ]Q x) \<union> ([ x \<leftarrow> m2 ] K (\<not> b) \<inter> [ x \<leftarrow> m2] Q x)"
  apply(auto simp add: Box_defs K_def split: sum.split sum.split_asm)
  apply(subgoal_tac "(\<lambda>y. True) = UNIV")
  apply(force)
  apply(force)
  apply(auto simp add: Box_defs)

lemma
  "([ x \<leftarrow> m ] K_set C x \<inter> [ x \<leftarrow> m ] P x) \<union> ([ x \<leftarrow> m ] K_set (Not o C) x \<inter> [x \<leftarrow> m] Q x) \<subseteq>
    [ x \<leftarrow> m ] (if C x then P x else Q x)"
  by(auto simp add: K_set_def o_def Box_defs split: sum.split sum.split_asm)

lemma BoxNN_if_then_elseP:
  "[ x \<leftarrow> m ] (if C x then P x else Q x) \<subseteq>
    ([ x \<leftarrow> m ] K_set C x \<inter> [ x \<leftarrow> m ] P x) \<union> ([ x \<leftarrow> m ] K_set (Not o C) x \<inter> [x \<leftarrow> m] Q x)"
  apply(auto)
  apply(clarsimp simp add: K_set_def o_def Box_defs split: sum.split sum.split_asm split_if_asm)
  apply(case_tac "C b")d
  by(auto simp add: K_set_def o_def Box_defs split: sum.split sum.split_asm)

lemma BoxNN_if_then_elseP:
  "[ x \<leftarrow> m ] (if C x then P x else Q x) = 
    ([ x \<leftarrow> m ] K_set C x \<inter> [ x \<leftarrow> m ] P x) \<union> ([ x \<leftarrow> m ] K_set (Not o C) x \<inter> [x \<leftarrow> m] Q x)"
  apply(auto simp add: K_set_def o_def Box_defs split: sum.split sum.split_asm)
  apply(simp add: mem_def)
  apply(
  apply(blast)
*)

lemma BoxNE_if_then_else: 
  "[@ x \<leftarrow> if b then m1 else m2] (Q x) = 
    (if b then [@ x \<leftarrow> m1] Q x else [@ x \<leftarrow> m2] Q x)"
by (simp)


primrec 
  wpcN_mmapR_rec:: 
    "('b list \<Rightarrow> 's set) => ('a \<Rightarrow> ('b, 's) r_monad) \<Rightarrow> 'b list \<Rightarrow> 'a list \<Rightarrow> 's set" where
  "wpcN_mmapR_rec P f l [] = (P l)"
  | "wpcN_mmapR_rec P f l (a#as) = BoxN (f a) (\<lambda>x. (wpcN_mmapR_rec P f (l@[x]) as))"

primrec 
  wpcE_mmapR_rec:: 
    "'s set \<Rightarrow> ('a \<Rightarrow> ('b, 's) r_monad) \<Rightarrow> 'a list \<Rightarrow> 's set" where
  "wpcE_mmapR_rec P f [] = true"
  | "wpcE_mmapR_rec P f (a#as) = 
     (BoxN (f a) (\<lambda>x. (wpcE_mmapR_rec P f as))) \<inter> (BoxR (f a) P)"

definition
  wpcN_mmapR :: "('b list \<Rightarrow> 's set) \<Rightarrow> ('a \<Rightarrow> ('b, 's) r_monad) \<Rightarrow> 'a list \<Rightarrow> 's set" where
  "wpcN_mmapR P f l \<equiv> wpcN_mmapR_rec P f [] l"

definition
  wpcE_mmapR :: "'s set \<Rightarrow> ('a \<Rightarrow> ('b, 's) r_monad) \<Rightarrow> 'a list \<Rightarrow> 's set" where
  "wpcE_mmapR P f l \<equiv> wpcE_mmapR_rec P f l"

declare wpcN_mmapR_def [simp] 
declare wpcE_mmapR_def [simp]

primrec 
  mmapR_rec :: "('a \<Rightarrow> ('b, 's) r_monad) \<Rightarrow> 'b list \<Rightarrow> 'a list \<Rightarrow> ('b list, 's) r_monad" where
  "mmapR_rec f l [] = returnR l"
  | "mmapR_rec f l (a#as) = (f a) \<guillemotright>=R (\<lambda>b. mmapR_rec f (l@[b]) as)"

definition
  mmapR :: "('a \<Rightarrow> ('b, 's) r_monad) \<Rightarrow> 'a list \<Rightarrow> ('b list, 's) r_monad" where
  "mmapR f l \<equiv> mmapR_rec f [] l"

lemma BoxN_mmapR_rec [rule_format, simp]: 
  "\<forall> l. [R ol \<leftarrow> mmapR_rec f l il] (P ol) = wpcN_mmapR_rec P f l il"
  by(induct il, simp_all add: BoxN_try_catch BoxN_return BoxN_bind)

lemma BoxN_mmapR [simp]: "[R ol \<leftarrow> mmapR f il] (P ol) = wpcN_mmapR P f il"
  by (simp add: mmapR_def)

lemma BoxR_mmapR_rec [rule_format, simp]: 
  "\<forall> l. [@R mmapR_rec f l il] P = wpcE_mmapR_rec P f il"
  by(induct il, simp_all add: BoxR_try_catch BoxR_return BoxR_bind)

lemma BoxR_mmapR [simp]: "[@R mmapR f il] P = wpcE_mmapR P f il"
by (simp add: mmapR_def)

lemma BoxN_empty: "[R x \<leftarrow> empty] P x = true"
  by(auto simp add: BoxN_def)

lemma BoxR_empty: "[@R empty] P = true"
  by(auto simp add: BoxR_def)

lemma BoxNN_empty: "[ x \<leftarrow> empty ] P x = true"
  by(auto simp add: BoxNN_def BoxN_empty)

lemma BoxNE_empty: "[@ x \<leftarrow> empty ] P x = true"
  by(auto simp add: BoxNE_def BoxN_empty)

lemma comp_eq_box_eq_inv: "(\<forall>s \<in> I. m s = m' s) = 
  ((\<forall>P. I \<inter> [R x \<leftarrow> m] (P x) = I \<inter> [R x \<leftarrow> m'] (P x)) \<and> (\<forall>Q. I \<inter> [@R m] Q = I \<inter> [@R m'] Q))" 
  (is "?L=?R")
proof
  assume ?L
  thus ?R 
    by(auto simp add: BoxN_def BoxR_def)
next
  assume A: ?R
  show ?L
  proof
    fix s
    assume inI: "s \<in> I"
    show "m s = m' s"
    proof(rule ccontr)
      assume not_eq: "m s \<noteq> m' s"
      def P \<equiv> "\<lambda>x. {s'. (Some x, s') \<in> m s}"
      def Q \<equiv> "{s'. (None, s') \<in> m s}"
      def P' \<equiv> "\<lambda>x. {s'. (Some x, s') \<in> m' s}"
      def Q' \<equiv> "{s'. (None, s') \<in> m' s}"
      have s_in: "s \<in> I \<inter> [R x \<leftarrow> m] (P x) \<and> s \<in> I \<inter> [@R m] Q \<and> 
        s \<in> I \<inter> [R x \<leftarrow> m'] (P' x) \<and> s \<in> I \<inter> [@R m'] Q'" 
        by(auto simp add: BoxN_def BoxR_def P_def Q_def P'_def Q'_def inI)
      have s_out: "s \<notin> I \<inter> [R x \<leftarrow> m'] (P x) \<or> s \<notin> I \<inter> [@R m'] Q \<or> 
        s \<notin> I \<inter> [R x \<leftarrow> m] (P' x) \<or> s \<notin> I \<inter> [@R m] Q'" 
        using not_eq
        apply(auto simp add: BoxN_def BoxR_def P_def Q_def P'_def Q'_def)
        apply(case_tac[!] a)
        apply(auto)
        done
      hence "\<exists>P Q. 
        (I \<inter> [R x \<leftarrow> m] (P x) \<noteq> I \<inter> [R x \<leftarrow> m'] (P x))
        \<or> (I \<inter> [@R m] Q \<noteq> I \<inter> [@R m'] Q)" 
        using s_in
        by(auto)
      thus False using A
        by(clarsimp)
    qed
  qed
qed

lemma comp_eq_box_eq: "(m = m') = 
  ((\<forall>P. [R x \<leftarrow> m] (P x) = [R x \<leftarrow> m'] (P x)) \<and> (\<forall>Q. [@R m] Q = [@R m'] Q))" 
  apply(insert comp_eq_box_eq_inv[where I=UNIV and m=m and m'=m', simplified])
  apply(simp only: fun_eq_iff)
  done

lemma Box_choice: "[N x \<leftarrow> m1 [+] m2] P x = [N x \<leftarrow> m1] P x \<inter> [N x \<leftarrow> m2]P x"
  by(auto simp add: choice_def Box_def)

lemma BoxN_choice: "[R x \<leftarrow> m1 [+] m2] P x = [R x \<leftarrow> m1] P x \<inter> [R x \<leftarrow> m2]P x"
  by(auto simp add: choice_def BoxN_def)

lemma BoxR_choice: "[@R m1 [+] m2] P = [@R m1] P \<inter> [@R m2]P"
  by(auto simp add: choice_def BoxR_def)

lemma BoxNN_choice: "[x \<leftarrow> m1 [+] m2] P x = [x \<leftarrow> m1] P x \<inter> [x \<leftarrow> m2] P x"
  by(auto simp add: choice_def Box_defs)

lemma BoxNE_choice: "[@ x \<leftarrow> m1 [+] m2] P x = [@ x \<leftarrow> m1] P x \<inter> [@ x \<leftarrow> m2] P x"
  by(auto simp add: choice_def Box_defs)

lemmas Box_simps =
    BoxNN_try_catch BoxNE_try_catch
    BoxNN_return BoxNE_return
    BoxNN_throw BoxNE_throw
    BoxNN_bind BoxNE_bind BoxNN_choice BoxNE_choice 
    BoxNN_if_then_else BoxNE_if_then_else 
    BoxNN_restart BoxNE_restart


end
