theory Persistency
imports EvaluationLogicR
begin

text {* We try to simplify the reasoning about programs with restarts
  to only those parts of the program which access the persistent
  storage. To that end, we partition the state into volatile and
  persistent components (denoted by type variables 'sv and 'sp in
  the following *}

type_synonym ('a, 'sv, 'sp)pr_monad = "('a, 'sv \<times> 'sp)r_monad"

definition liftP :: "('a, 'sv)r_monad \<Rightarrow> ('a, 'sv, 'sp)pr_monad" where
  "liftP m \<equiv> \<lambda>(sv, sp). (\<lambda>(x, sv'). (x, sv', sp)) ` m sv"

text {* We want to boil down all our reasoning to the following form
  of programs:

  @{term "
  <tryR>
    computation
  <catchR> restart_handler
  "}
  
  and then only care about normal termination of such computations
  (i.e. not care about restarts within the restart
  handler). Intuitively, the behaviour of the restart handler should
  not depend on the state of the volatile memory; formally, any
  predicate P, the precondition @{term "[R x \<leftarrow> restart_handler] P"}
  should not depend on the volatile part of the storage. 

  We try to capture that notion with the following definition of
  "volatile memory-invariant" predicates.
    
*}

definition vol_invP :: "('sv \<times> 'sp)set \<Rightarrow> bool" where
  "vol_invP P \<equiv> \<exists>Pp. P = UNIV \<times> Pp"

lemma vol_invP_alter:
  "vol_invP P = (\<forall>sp sv sv'. (sv, sp) \<in> P \<longrightarrow> (sv', sp) \<in> P)"
proof
  assume "vol_invP P"
  thus "(\<forall>sp sv sv'. (sv, sp) \<in> P \<longrightarrow> (sv', sp) \<in> P)" 
    by(auto simp add: vol_invP_def)
next
  assume "(\<forall>sp sv sv'. (sv, sp) \<in> P \<longrightarrow> (sv', sp) \<in> P)" 
  hence "P = UNIV \<times> snd ` P"
    by force
  thus "vol_invP P"
    by(unfold vol_invP_def, blast)
qed

definition vol_inv :: "('a, 'sv, 'sp)pr_monad \<Rightarrow> bool" where
  "vol_inv m \<equiv> (\<forall>P. vol_invP ([R x \<leftarrow> m] P)) \<and> (\<forall>Q. vol_invP ([@R m] Q))"

text {* The definition we will mostly use in proofs *}

lemma vol_inv_alter: "vol_inv m = (
  (\<forall>P sp sv sv'. (sv, sp) \<in> [R x \<leftarrow> m] P \<longrightarrow> (sv', sp) \<in> [R x \<leftarrow> m] P)
  \<and> (\<forall>Q sp sv sv'. (sv, sp) \<in> [@R m]Q \<longrightarrow> (sv', sp) \<in> [@R m]Q))"
  by(auto simp add: vol_inv_def vol_invP_alter)

end
