theory PmmDefs
imports Cartesian_Equipotent Infinite RMonad Card_Mode (* EvaluationLogicR *)
begin

ML_file "monad_defs.ML"

attribute_setup monad_def =
  {* Attrib.add_del Monad_Defs.add Monad_Defs.del *}
  "definitions of composed monadic operations"

(*************************************************************)
subsection {* Indices *}
(*************************************************************)
datatype admin_indx = HASH_REG | PTR_REG

typedecl pub_indx 

datatype indx = Public pub_indx | Admin admin_indx

axiomatization where
  UNIV_pindx: 
  "finite (UNIV :: pub_indx set)"

lemma
  UNIV_aindx:
  "finite (UNIV :: admin_indx set)"
proof-
  have eq: "(UNIV :: admin_indx set) = {HASH_REG, PTR_REG}"
    apply(rule UNIV_eq_I)
    apply(case_tac x)
    apply(auto)
    done

  moreover
  have fin: "finite {HASH_REG, PTR_REG}"
    by(auto)

  finally show ?thesis .
qed

lemma UNIV_indx:
  "finite (UNIV :: indx set)"
proof-
  have "(UNIV :: indx set) = Admin ` UNIV \<union> Public ` UNIV"
    apply(auto)
    apply(case_tac x, auto)
    done
  moreover have "finite (Admin ` UNIV \<union> Public ` UNIV)"
    by(auto simp add: UNIV_pindx UNIV_aindx)
  ultimately show ?thesis
    by(auto)
qed
    
instance indx :: finite proof
qed (rule UNIV_indx)

instance pub_indx :: finite proof
qed (rule UNIV_pindx)

instance admin_indx :: finite proof
qed (rule UNIV_aindx)

(*************************************************************)
subsection {* Instances *}
(*************************************************************)

datatype INST = Primary | Secondary

lemma[iff]: "(x \<noteq> Primary) = (x = Secondary)"
            "(Primary \<noteq> x) = (Secondary = x)"
            "(x \<noteq> Secondary) = (x = Primary)"
            "(Secondary \<noteq> x) = (Primary = x)"
  by(case_tac[!] x, auto)

lemma UNIV_inst: "(UNIV :: INST set) = {Primary, Secondary}"
  by(auto)

instance INST :: finite proof
qed (simp add: UNIV_inst)

primrec other_i :: "INST \<Rightarrow> INST" where
  "other_i Primary = Secondary"
  | "other_i Secondary = Primary"

lemma other_i_inst[iff]: "other_i inst \<noteq> inst"
  by(case_tac inst, simp_all)

(*************************************************************)
subsection {* Exceptions *}
(*************************************************************)

text {*
  Degraded exceptions can carry values. The types of values they
  carry depend on where they are used, and can be either unit (for startup), boolean 
  (for update) or content type values (for fetch).
*}
datatype 'c exc_val = EU | EB bool | EC 'c

datatype 'c exceptions =
  FailWrite | FailLLRead | FailRead | FailMode | Degraded "'c exc_val" | InvalidInput | StartedExc

fun get_exc_val :: "'c exceptions \<Rightarrow> 'c exc_val" where
  "get_exc_val (Degraded v) = v"

fun get_exc_val_cnt :: "'c exc_val \<Rightarrow> 'c" where
  "get_exc_val_cnt (EC cnt) = cnt"

(*************************************************************)
subsection {* PMM restart-exception monad *}
(*************************************************************)

type_synonym ('a, 'e, 'sv, 'sp)pre_monad = "('a, 'e, 'sv \<times> 'sp)re_monad"

(*************************************************************)
subsection {* Hashes and capacities *}
(*************************************************************)

class all_sizes = size +
  assumes all_sizes: "\<forall>n. \<exists>x :: 'a. size x = n"

datatype 'a hash = Hash 'a

datatype 'a htable = HTable "pub_indx \<times> INST \<Rightarrow> 'a hash" "INST \<Rightarrow> 'a hash"

fun update_htable where
  "update_htable ht (Admin HASH_REG) _ _ = ht"
  | "update_htable (HTable pub_h ptr_h) (Admin PTR_REG) inst cnt = 
        HTable pub_h (ptr_h (inst := Hash cnt))"
  | "update_htable (HTable pub_h ptr_h) (Public pind) inst cnt = 
        HTable (pub_h((pind, inst) := Hash cnt)) ptr_h"

fun pub_hashes :: "'c htable \<Rightarrow> pub_indx \<times> INST \<Rightarrow> 'c hash" where
  "pub_hashes (HTable pub_ht _) = pub_ht"

fun ptr_hashes :: "'c htable \<Rightarrow> INST \<Rightarrow> 'c hash" where
  "ptr_hashes (HTable _ ptr_ht) = ptr_ht"

definition matches_htable :: "'c htable \<Rightarrow> indx \<Rightarrow> INST \<Rightarrow> 'c \<Rightarrow> bool" where
  "matches_htable ht ind inst cnt \<equiv>
    (ind = Admin PTR_REG \<longrightarrow> Hash cnt = ptr_hashes ht inst)
    \<and> (\<forall>pind. ind = Public pind \<longrightarrow> Hash cnt = pub_hashes ht (pind, inst))
  "

consts HTABLE_SIZE :: nat
consts PTR_SIZE :: nat

consts capacity :: "indx \<Rightarrow> nat"

specification (capacity)
  htable_size: "HTABLE_SIZE <= capacity (Admin HASH_REG)"
  ptr_size: "PTR_SIZE <= capacity (Admin PTR_REG)"
proof-
  def cap_f \<equiv> "\<lambda>ind. if ind = Admin HASH_REG 
                    then HTABLE_SIZE 
                    else if ind = Admin PTR_REG
                        then PTR_SIZE
                        else undefined"
  show ?thesis 
    apply(rule exI[where x=cap_f])
    apply(simp add: cap_f_def)
    done
qed

class serialization_cl = cartesian_equipotent + infinite

consts parse_htable :: "('a :: serialization_cl) \<Rightarrow> 'a htable"
consts serialize_htable :: "'a htable \<Rightarrow> ('a :: serialization_cl)"
consts parse_hinst :: "('a::serialization_cl) \<Rightarrow> 'a \<times> 'a hash"
consts serialize_hinst :: "('a::serialization_cl) \<times> 'a hash \<Rightarrow> 'a"

consts parse_ptr :: "'a :: serialization_cl \<Rightarrow> pub_indx set"
consts serialize_ptr :: "pub_indx set \<Rightarrow> 'a :: serialization_cl"

definition snd_to_hash :: "'a \<times> 'a \<Rightarrow> 'a \<times> 'a hash" where
  "snd_to_hash p \<equiv> (fst p, Hash (snd p))"

lemma bij_snd_to_hash: "bij snd_to_hash"
  apply(auto simp add: bij_def inj_on_def snd_to_hash_def)
  apply(case_tac "b")
  apply(auto simp add: snd_to_hash_def)
  apply(rule_tac x="(a, aa)" in range_eqI)
  apply(auto simp add: snd_to_hash_def)
  done

specification(serialize_hinst)
  bij_serialize_hinst: 
  "bij serialize_hinst"
proof-
  from map_bij obtain f :: "'a \<Rightarrow> 'a \<times> 'a" where "bij f" ..
  hence "\<exists>(g :: 'a \<Rightarrow> 'a \<times> 'a hash). bij g" (is "\<exists>g. ?P g") using bij_snd_to_hash
    by(auto intro!: bij_comp)
  then obtain g :: "'a \<Rightarrow> 'a \<times> 'a hash" where "bij g" ..
  from this[THEN bij_imp_bij_inv] show ?thesis by auto
qed

specification(parse_hinst)
 parse_hinst_spec: "parse_hinst = inv (serialize_hinst)"
  by auto

lemma bij_Hash: "bij Hash"
  apply(auto simp add: bij_def inj_on_def surj_def)
  apply(case_tac y, auto)
  done

lemma inj_comp_curried: "inj f \<Longrightarrow> inj (comp f)"
proof-
  assume inj_f: "inj f"
  show ?thesis
  proof(rule injI, rule ccontr)
    fix g h
    assume neq: "g \<noteq> h" and eq: "f o g = f o h"
    from neq obtain x where "g x \<noteq> h x"
      by(metis ext)
    hence "f o g \<noteq> f o h" using inj_f
      by (metis comp_apply the_inv_f_f)
    thus False using eq ..
  qed
qed

lemma bij_comp_curried: "bij f \<Longrightarrow> bij (comp f)"
proof-
  assume bij_f: "bij f"
  have inj: "inj (comp f)"
    by(intro inj_comp_curried bij_is_inj bij_f)
  moreover from bij_f have surj: "surj (comp f)"
    apply(auto simp add: surj_def)
    apply(rule_tac x="inv f o y" in exI)
    apply(auto simp add: bij_def o_assoc surj_iff)
    done
  ultimately show ?thesis
    by(auto simp add: bij_def)
qed

datatype serialize_aux = IAux INST | PAux "pub_indx \<times> INST"

lemma UNIV_serialize_aux: "finite (UNIV :: serialize_aux set)"
proof-
  have "(UNIV :: serialize_aux set) = IAux ` UNIV \<union> PAux ` UNIV"
    apply(auto)
    apply(case_tac x, auto)
    done
  moreover have "finite (IAux ` UNIV \<union> PAux ` UNIV)"
    by(auto)
  ultimately show ?thesis
    by(auto)
qed

instance serialize_aux :: finite proof qed (rule UNIV_serialize_aux)

specification (serialize_htable)
 inj_serialize_htable: "inj serialize_htable" 
proof-
  have "\<exists>f :: 'a htable \<Rightarrow> 'a. bij f"
  proof-
    have "\<exists>f :: 'a htable \<Rightarrow> (serialize_aux \<Rightarrow> 'a hash). bij f"
    proof-
      def g \<equiv> "\<lambda>(ht :: 'a htable) sa. 
            case sa of IAux i \<Rightarrow> ptr_hashes ht i | PAux pii \<Rightarrow> pub_hashes ht pii"
      have "inj g"
        apply(rule inj_onI)
        apply(case_tac x, case_tac y, auto simp add: g_def intro!: ext)
        apply (metis serialize_aux.simps(6))
        apply (metis serialize_aux.simps(5))
        done
      have "surj g"
      proof(rule surjI)
        fix x
        def f \<equiv> "\<lambda>h :: serialize_aux \<Rightarrow> 'a hash. HTable (h o PAux) (h o IAux)"
        show "g (f x) = x"
          apply(rule ext)
          apply(case_tac xa)
          apply(auto simp add: surj_def g_def f_def)
          done
      qed
      from `inj g` `surj g` have "bij g" by (rule bijI)
      thus ?thesis by auto
    qed
    then obtain f :: "'a htable \<Rightarrow> (serialize_aux \<Rightarrow> 'a hash)" where bij_f: "bij f" ..
    from bij_finite_maps_rev obtain g :: "(serialize_aux \<Rightarrow> 'a) \<Rightarrow> 'a" where "bij g" ..
    moreover from bij_Hash have "bij (inv Hash)" by(rule bij_imp_bij_inv)
    ultimately show ?thesis using bij_comp_curried bij_f
      apply(rule_tac x="(g o (comp (inv Hash))) o f" in exI)
      apply(auto intro: bij_comp)
      done
  qed
  thus ?thesis
    by(auto dest: bij_is_inj)
qed

specification (parse_htable)
 parse_htable_spec: "parse_htable = inv (serialize_htable)" 
  by auto

specification (serialize_ptr)
  inj_serialize_ptr: "inj serialize_ptr"
  by(rule fin_to_inf_inj)

specification (parse_ptr)
  parse_ptr_spec: "parse_ptr = inv (serialize_ptr)"
  by(auto)

class serialization2_cl = serialization_cl + all_sizes +
  fixes 
  htable_to_hinst_cnt :: "'a htable \<Rightarrow> 'a"

class contents_cl = serialization2_cl +
  assumes 
  htable_to_hinst_cnt_def: "(htable_to_hinst_cnt :: 'a htable \<Rightarrow> 'a :: serialization2_cl)  ht = 
    serialize_hinst (serialize_htable ht, Hash (serialize_htable ht))"
  and size_serialize_hashes: "size (htable_to_hinst_cnt ht) = HTABLE_SIZE"
  and size_serialize_ptr: "size (serialize_ptr pinds) = PTR_SIZE"

type_synonym 'c pub_hashes = "(pub_indx \<times> INST) \<Rightarrow> 'c hash"

(*************************************************************)
subsection {* Miscellaneous *}
(*************************************************************)

datatype sync_direction = SEq | SR | SL | SBot

abbreviation fun_upd_second :: "('a \<Rightarrow> 'b \<Rightarrow> 'c) \<Rightarrow> 'a \<Rightarrow> 'b \<Rightarrow> 'c \<Rightarrow> 'a \<Rightarrow> 'b \<Rightarrow> 'c" where
  "fun_upd_second f x y r \<equiv> f(x := (f x)(y := r))"

abbreviation fun_override_on2 :: 
  "('a \<Rightarrow> 'b \<Rightarrow> 'c) \<Rightarrow>  ('a \<Rightarrow> 'b \<Rightarrow> 'c) \<Rightarrow> 'a set \<Rightarrow> 'b set \<Rightarrow> 'a \<Rightarrow> 'b => 'c"
  where
  "fun_override_on2 f g X Y \<equiv> (\<lambda>a b. if a \<in> X \<and> b \<in> Y then g a b else f a b)"

(*************************************************************)
subsection {* Volatile memory *}
(*************************************************************)

definition getsV :: "('sv \<Rightarrow> 'a) \<Rightarrow> ('a, 'e, 'sv, 'sp)pre_monad" where
  "getsV f \<equiv> getsRE (f o fst)"

definition modifyV :: "('sv \<Rightarrow> 'sv) \<Rightarrow> (unit, 'e, 'sv, 'sp)pre_monad" where
  "modifyV f \<equiv> do { (sv, sp) \<leftarrow> getsRE id; putRE (f sv, sp) }"

definition putV :: "'sv \<Rightarrow> (unit, 'e, 'sv, 'sp)pre_monad" where
  "putV sv' \<equiv> do { (sv, sp) \<leftarrow> getsRE id; putRE (sv', sp) }"

definition restart_mangle_sv :: "('a, 'sv \<times> 'sp)r_monad" where
  "restart_mangle_sv \<equiv> doN {
    sv' \<leftarrow> choose UNIV;
    modify (\<lambda>(sv, sp). (sv', sp));
    restart
  }"

lemma restart_mangle_sv_raw:
  "restart_mangle_sv = (\<lambda>(sv, sp). {(None, sv', sp)|sv'. True})"
  by(auto simp add: restart_mangle_sv_def unfold_monad choose_def modify_def gets_def put_def put'_def)

record pmm_vol_mem =
  card_mode :: card_mode
  updated :: "pub_indx set"

definition get_mode :: "(card_mode, 'e, 'a pmm_vol_mem_scheme \<times> 'sp)re_monad" where
  "get_mode = getsV card_mode"

definition set_degraded :: "(unit, 'e, 'a pmm_vol_mem_scheme \<times> 'sp)re_monad" where
  "set_degraded \<equiv> modifyV (\<lambda>s. s\<lparr> card_mode := min CDegraded (card_mode s) \<rparr> )"

definition set_degraded_p :: "(unit, 'e, 'a pmm_vol_mem_scheme \<times> 'sp)re_monad" where
  "set_degraded_p \<equiv> modifyV (\<lambda>s. s\<lparr> card_mode := min CDegradedP (card_mode s) \<rparr> )"

definition set_degraded_h :: "(unit, 'e, 'a pmm_vol_mem_scheme \<times> 'sp)re_monad" where
  "set_degraded_h \<equiv> modifyV (\<lambda>s. s\<lparr> card_mode := min CDegradedH (card_mode s) \<rparr> )"

definition set_failed :: "(unit, 'e, 'a pmm_vol_mem_scheme \<times> 'sp)re_monad" where
  "set_failed \<equiv> modifyV (\<lambda>s. s\<lparr> card_mode := CFail \<rparr> )"

definition get_updated :: "(pub_indx set, 'e, 'a pmm_vol_mem_scheme \<times> 'sp)re_monad" where
  "get_updated \<equiv> getsV updated"

definition set_updated :: "pub_indx set \<Rightarrow> (unit, 'e, 'a pmm_vol_mem_scheme \<times> 'sp)re_monad" where
  "set_updated val \<equiv> modifyV (\<lambda>s. s\<lparr> updated := val \<rparr> )"

definition add_to_updated :: "pub_indx \<Rightarrow> (unit, 'e, 'a pmm_vol_mem_scheme \<times> 'sp)re_monad" where
  "add_to_updated pind \<equiv> do {
    updated \<leftarrow> get_updated;
    set_updated (insert pind updated)
  }"

definition remove_from_updated :: "pub_indx \<Rightarrow> (unit, 'e, 'a pmm_vol_mem_scheme \<times> 'sp)re_monad" where
  "remove_from_updated pind \<equiv> do {
    updated \<leftarrow> get_updated;
    set_updated (updated - {pind})
  }"

lemmas getters_setters =
  modifyV_def getsV_def getsRE_def getsR_def gets_def putRE_def putRE'_def putR'_def put'_def

definition init_vars :: "(unit, 'e, 'a pmm_vol_mem_scheme, 'sp)pre_monad" where
  init_vars_def[monad_def]:
  "init_vars \<equiv> modifyV (\<lambda>s. s\<lparr> 
    card_mode := CNormal,
    updated := {}
  \<rparr> )"

record abs_vol_mem =
  abs_card_mode :: abs_card_mode
  abs_updated :: "pub_indx set"

definition get_abs_mode :: "(abs_card_mode, 'e, 'a abs_vol_mem_scheme \<times> 'sp)re_monad" where
  "get_abs_mode = getsV abs_card_mode"

definition set_abs_degraded :: "(unit, 'e, 'a abs_vol_mem_scheme \<times> 'sp)re_monad" where
  "set_abs_degraded \<equiv> modifyV (\<lambda>s. s\<lparr> abs_card_mode := min ACDegraded (abs_card_mode s) \<rparr> )"

definition set_abs_failed :: "(unit, 'e, 'a abs_vol_mem_scheme \<times> 'sp)re_monad" where
  "set_abs_failed \<equiv> modifyV (\<lambda>s. s\<lparr> abs_card_mode := ACFail \<rparr> )"

definition get_abs_updated :: "(pub_indx set, 'e, 'a abs_vol_mem_scheme \<times> 'sp)re_monad" where
  "get_abs_updated \<equiv> getsV abs_updated"

definition set_abs_updated :: "pub_indx set \<Rightarrow> (unit, 'e, 'a abs_vol_mem_scheme \<times> 'sp)re_monad" where
  "set_abs_updated val \<equiv> modifyV (\<lambda>s. s\<lparr> abs_updated := val \<rparr> )"

definition add_to_abs_updated :: "pub_indx \<Rightarrow> (unit, 'e, 'a abs_vol_mem_scheme \<times> 'sp)re_monad" where
  "add_to_abs_updated pind \<equiv> do {
    abs_updated \<leftarrow> get_abs_updated;
    set_abs_updated (insert pind abs_updated)
  }"

definition remove_from_abs_updated :: "pub_indx \<Rightarrow> (unit, 'e, 'a abs_vol_mem_scheme \<times> 'sp)re_monad" where
  "remove_from_abs_updated pind \<equiv> do {
    abs_updated \<leftarrow> get_abs_updated;
    set_abs_updated (abs_updated - {pind})
  }"

definition init_abs_vars :: "(unit, 'e, 'a abs_vol_mem_scheme, 'sp)pre_monad" where
  init_abs_vars_def[monad_def]:
  "init_abs_vars \<equiv> modifyV (\<lambda>s. s\<lparr> 
    abs_card_mode := ACNormal,
    abs_updated := {}
  \<rparr> )"

lemmas vol_ops_defs =
  get_mode_def
  set_degraded_def
  set_degraded_p_def
  set_degraded_h_def
  set_failed_def
  get_updated_def
  set_updated_def
  add_to_updated_def
  remove_from_updated_def
  init_vars_def
  get_abs_mode_def
  set_abs_degraded_def
  set_abs_failed_def
  get_abs_updated_def
  set_abs_updated_def
  add_to_abs_updated_def
  remove_from_abs_updated_def
  init_abs_vars_def

(*************************************************************)
subsection {* Definitions common to multiple models *}
(*************************************************************)

text {*
  Using ifs instead of case here for reasons of laziness (not having
  to adapt pmm_tac to handle @{typ card_mode} cases
  *}

definition throw_mode_error 
  :: "'c exc_val \<Rightarrow> (unit, 'c exceptions, 'a pmm_vol_mem_scheme, 'sp)pre_monad" 
  where throw_mode_error_def[monad_def]:
  "throw_mode_error val \<equiv> do {
    mode \<leftarrow> get_mode;
    if mode = CFail
    then throw FailMode
    else if mode \<in> {CDegradedH, CDegradedP, CDegraded}
    then throw (Degraded val)
    else return ()
  }"

definition throw_abs_mode_error 
  :: "'c exc_val \<Rightarrow> (unit, 'c exceptions, 'a abs_vol_mem_scheme, 'sp)pre_monad" 
  where throw_abs_mode_error_def[monad_def]:
  "throw_abs_mode_error val \<equiv> do {
    mode \<leftarrow> get_abs_mode;
    if mode = ACFail
    then throw FailMode
    else if mode = ACDegraded
    then throw (Degraded val)
    else return ()
  }"

definition sv_mode_equals :: "card_mode \<Rightarrow> ('a pmm_vol_mem_scheme \<times> 'sp)set" where 
  "sv_mode_equals m \<equiv> {(sv, sp). card_mode sv = m}"

definition sv_mode_neq :: "card_mode \<Rightarrow> ('a pmm_vol_mem_scheme \<times> 's)set" where
  "sv_mode_neq m \<equiv> {(sv, sp). card_mode sv \<noteq> m}"

definition sv_abs_mode_equals :: "abs_card_mode \<Rightarrow> ('a abs_vol_mem_scheme \<times> 'sp)set" where 
  "sv_abs_mode_equals m \<equiv> {(sv, sp). abs_card_mode sv = m}"

definition sv_abs_mode_neq :: "abs_card_mode \<Rightarrow> ('a abs_vol_mem_scheme \<times> 's)set" where
  "sv_abs_mode_neq m \<equiv> {(sv, sp). abs_card_mode sv \<noteq> m}"

definition updated_equals where
  "updated_equals pinds \<equiv> {(sv, sp). updated sv = pinds}"

definition abs_updated_equals where
  "abs_updated_equals pinds \<equiv> {(sv, sp). abs_updated sv = pinds}"

definition fun_map_override_on :: "('a \<Rightarrow> 'b) \<Rightarrow> ('a ~=> 'b) \<Rightarrow> 'a set \<Rightarrow> ('a \<Rightarrow> 'b)" where
  "fun_map_override_on f g A = override_on f (\<lambda>x. if x \<in> dom g then the (g x) else f x) A"

definition check_sizes where
  check_sizes_def[monad_def]:
  "check_sizes upd_f \<equiv> do {
    ord \<leftarrow> chooseN (order_set (dom upd_f));
    mapM' (\<lambda>pind. 
        if (size o the o upd_f) pind > capacity (Public pind)
        then throw InvalidInput
        else return ()
    ) ord
  }"

end

