theory Specification
imports Main
begin

record 's TS =
  init :: "'s set"
  init_trans :: "('s \<times> 's)set" (* a distinguished initial transition *)
  trans :: "('s \<times> 's)set"

text {* The inductive set of reachable states. Note that the initial
states are not considered reachable in this definition; we first perform
the initial transition *}

inductive_set 
  reach ::  "('s, 'a) TS_scheme \<Rightarrow> 's set" 
  for T :: "('s, 'a) TS_scheme"
  where
  r_init [intro]:  "s \<in> init_trans T `` init T \<Longrightarrow> s \<in> reach T"
  | r_trans [intro]: "\<lbrakk> (s, t) \<in> trans T; s \<in> reach T \<rbrakk> \<Longrightarrow> t \<in> reach T"

inductive_set 
  beh :: "('s, 'a) TS_scheme \<Rightarrow> ('s list) set"
  for T :: "('s, 'a) TS_scheme" 
  where
  b_empty [iff]: "[] \<in> beh T"
  | b_init [intro]: "s \<in> init_trans T `` init T \<Longrightarrow> [s] \<in> beh T"
  | b_trans [intro]: "\<lbrakk> s # b \<in> beh T; (s, t) \<in> trans T \<rbrakk> \<Longrightarrow> t # s # b \<in> beh T"

inductive_cases beh_non_empty: "s # b \<in> beh T"

thm beh_non_empty

lemma beh_immediate_prefix_closed:
  "s # b \<in> beh T \<Longrightarrow> b \<in> beh T"
  by (erule beh_non_empty, auto) 

lemma beh_prefix_closed:
  "c @ b \<in> beh T \<Longrightarrow> b \<in> beh T"
  by (induct c, auto dest!: beh_immediate_prefix_closed)

lemma beh_in_reach [rule_format]:
  "b \<in> beh T \<Longrightarrow> (\<forall>s \<in> set b. s \<in> reach T)"
  by (erule beh.induct) (auto)

lemma reach_in_beh:
  "s \<in> reach T \<Longrightarrow> \<exists>b \<in> beh T. s \<in> set b"
  apply (erule reach.induct, auto)
  apply (rule_tac x="[s]" in bexI, auto)

  apply (drule split_list, clarsimp)
  apply (rule_tac x="t # s # zs" in bexI, 
    auto intro: beh_prefix_closed)
  done

lemma reach_equiv_beh_states: "reach T = \<Union> set`(beh T)"
  by (auto intro!: reach_in_beh beh_in_reach)

record ('s, 'o) spec = "'s TS" +
  obs :: "'s \<Rightarrow> 'o" 

lemma beh_obs_upd [simp]: "beh (S(| obs := x |)) = beh S"
  by (safe) (erule_tac[!] beh.induct, auto intro!: b_init)

lemma reach_obs_upd [simp]: "reach (S(| obs := x |)) = reach S"
  apply (safe) 
  apply (erule_tac[!] reach.induct, auto)
  apply(auto intro!: r_init)
  done

text {* Observable behaviour and reachability. *}

definition 
  obeh :: "('s, 'o) spec \<Rightarrow> ('o list) set" where
  "obeh S \<equiv> (map (obs S))`(beh S)"

definition 
  oreach :: "('s, 'o) spec \<Rightarrow> 'o set" where 
  "oreach S \<equiv> (obs S)`(reach S)"

lemma oreach_equiv_obeh_states:
  "oreach S = \<Union> set`(obeh S)"
  by (auto simp add: reach_equiv_beh_states oreach_def obeh_def)

lemma reach_spec:
  "reach (TS.extend t ob) = reach t"
  apply(auto simp add: TS.extend_def)
  apply(erule reach.induct, auto)
  apply(erule reach.induct, auto)
  apply(auto intro!: reach.r_init)
  done

lemma obs_ts:
  "obs (TS.extend t (spec.fields f)) = f"
  by(auto simp add: TS.extend_def spec.fields_def)

lemma obeh_pi_translation:
  "(map pi)`(obeh S) = obeh (S(| obs := pi o (obs S) |))"
  by(force simp add: obeh_def)

lemma oreach_pi_translation:
  "pi`(oreach S) = oreach (S(| obs := pi o (obs S) |))"
  by (auto simp add: oreach_def)

definition 
  implements :: "['p \<Rightarrow> 'o, ('s, 'o) spec, ('t, 'p) spec] \<Rightarrow> bool" where
  "implements pi Sa Sc \<equiv> (map pi)`(obeh Sc) \<subseteq> obeh Sa"

text {* Reflexivity and transitivity *}

lemma implements_refl: "implements id S S"
  by (auto simp add: implements_def)

lemma implements_trans:
  "\<lbrakk> implements pi1 S1 S2; implements pi2 S2 S3 \<rbrakk> 
  \<Longrightarrow> implements (pi1 o pi2) S1 S3"
  by(force simp add: implements_def)

text {* Preservation of external invariants *}

lemma implements_oreach:
  "implements pi Sa Sc \<Longrightarrow> pi`(oreach Sc) \<subseteq> oreach Sa"
  by (auto simp add: implements_def oreach_equiv_obeh_states dest!: subsetD)

lemma external_invariant_preservation:
  "\<lbrakk> oreach Sa \<subseteq> Q; implements pi Sa Sc \<rbrakk>
  \<Longrightarrow> pi`(oreach Sc) \<subseteq> Q"
  by (rule subset_trans [OF implements_oreach]) (auto)

definition
  PO_rhoare :: 
  "[('s \<times> 't) set, ('s \<times> 's) set, ('t \<times> 't) set, ('s \<times> 't) set] \<Rightarrow> bool"
("(4{_} _, _ {> _})" [0, 0, 0] 90)
  where
  "{pre} Ra, Rc {> post} \<equiv> pre O Rc \<subseteq> Ra O post"

lemmas PO_rhoare_defs = PO_rhoare_def relcomp_unfold

lemma relhoare_conseq_left [intro]:
  "\<lbrakk> {pre'} Ra, Rc {> post}; pre \<subseteq> pre' \<rbrakk> 
  \<Longrightarrow> {pre} Ra, Rc {> post}"
  by (auto simp add: PO_rhoare_defs dest!: subsetD)

lemma relhoare_conseq_right:                    -- {* do NOT declare [intro] *}
  "\<lbrakk> {pre} Ra, Rc {> post'}; post' \<subseteq> post \<rbrakk> 
  \<Longrightarrow> {pre} Ra, Rc {> post}"
  by (auto simp add: PO_rhoare_defs)

lemma relhoare_false_left [simp]:               -- {* do NOT declare [intro] *}
  "{ {} } Ra, Rc {> post}"
  by (auto simp add: PO_rhoare_defs)

lemma relhoare_true_right [simp]:                -- {* not true in general *}
  "{pre} Ra, Rc {> UNIV} = (Domain (pre O Rc) \<subseteq> Domain Ra)"
  by (auto simp add: PO_rhoare_defs)

lemma Domain_rel_comp [intro]:
  "Domain pre \<subseteq> R \<Longrightarrow> Domain (pre O Rc) \<subseteq> R"
  by (auto simp add: Domain_def)

lemma rel_hoare_skip [iff]: "{R} Id, Id {> R}"
  by (auto simp add: PO_rhoare_def)


text {* Reflexivity and transitivity. *}

lemma relhoare_refl [simp]: "{Id} R, R {> Id}"
  by (auto simp add: PO_rhoare_defs)

lemma rhoare_trans:
  "\<lbrakk> {R1} T1, T2 {> R1}; {R2} T2, T3 {> R2} \<rbrakk>
  \<Longrightarrow> {R1 O R2} T1, T3 {> R1 O R2}"
  apply (auto simp add: PO_rhoare_def del: subsetI)
  apply (drule subset_refl [THEN relcomp_mono, where r=R1])
  apply (drule subset_refl [THEN [2] relcomp_mono, where s=R2])
  apply (auto simp add: O_assoc del: subsetI)
  done


text {* Conjunction in the post-relation cannot be split in general.  However, 
  here are two useful special cases.  In the first case the abstract transtition 
  relation is deterministic and in the second case one conjunct is a cartesian 
  product of two state predicates.  *}

lemma relhoare_conj_right_det:                 
  "\<lbrakk> {pre} Ra, Rc {> post1}; {pre} Ra, Rc {> post2};
  single_valued Ra \<rbrakk>                           (* only for deterministic Ra! *)  
  \<Longrightarrow> {pre} Ra, Rc {> post1 \<inter> post2}"
  by (auto simp add: PO_rhoare_defs dest: single_valuedD dest!: subsetD)

text {* Separate rule for cartesian products. *}

lemma relhoare_concrete_Un:
  "{pre} Ra, Rc1 \<union> Rc2 {> post} 
  = ({pre} Ra, Rc1 {> post} \<and> {pre} Ra, Rc2 {> post})"
  apply (auto simp add: PO_rhoare_defs)
  apply (auto dest!: subsetD)
  done

lemma relhoare_concrete_UN:
  "{pre} Ra, \<Union>x. Rc x {> post} = (\<forall>x. {pre} Ra, Rc x {> post})"
apply (auto simp add: PO_rhoare_defs)
apply (auto dest!: subsetD)
done

lemma relhoare_abstract_Un_left:
  "\<lbrakk> {pre} Ra1, Rc {> post} \<rbrakk>
  \<Longrightarrow> {pre} Ra1 \<union> Ra2, Rc {> post}"
  by (auto simp add: PO_rhoare_def)

lemma relhoare_abstract_Un_right:
  "\<lbrakk> {pre} Ra2, Rc {> post} \<rbrakk>
  \<Longrightarrow> {pre} Ra1 \<union> Ra2, Rc {> post}"
  by (auto simp add: PO_rhoare_def)

lemma relhoare_abstract_UN:
  "\<lbrakk> {pre} Ra x, Rc {> post} \<rbrakk>
  \<Longrightarrow> {pre} \<Union>x. Ra x, Rc {> post}"
  apply (auto simp add: PO_rhoare_def)
  apply (auto dest!: subsetD)
  done

lemma relhoare_UN_lockstep:
  "\<lbrakk> \<forall>x. {pre} Ra x, Rc x {> post} \<rbrakk>
  \<Longrightarrow> {pre} \<Union>x. Ra x, \<Union>x. Rc x {> post}"
  apply (auto simp add: PO_rhoare_def)
  apply (force dest!: subsetD)
  done

definition PO_refines 
  :: "[('s \<times> 't) set, ('s, 'a) TS_scheme, ('t, 'b) TS_scheme] \<Rightarrow> bool" 
  where
  "PO_refines R Ta Tc \<equiv> (
    init_trans Tc `` init Tc \<subseteq> R``(init_trans Ta `` init Ta)
    \<and> {R} (trans Ta), (trans Tc) {> R} 
  )"

lemma refine_basic:
  "\<lbrakk> init_trans Tc `` init Tc \<subseteq> R``(init_trans Ta `` init Ta); {R} (trans Ta), (trans Tc) {> R} \<rbrakk>
  \<Longrightarrow> PO_refines R Ta Tc"
  by (simp add: PO_refines_def)

inductive_set 
  seq_lift :: "('s \<times> 't) set \<Rightarrow> ('s list \<times> 't list) set" 
  for R :: "('s \<times> 't) set"
where
  sl_nil [iff]: "([], []) \<in> seq_lift R"
| sl_cons [intro]: 
    "\<lbrakk> (xs, ys) \<in> seq_lift R; (x, y) \<in> R \<rbrakk> \<Longrightarrow> (x#xs, y#ys) \<in> seq_lift R"

inductive_cases sl_cons_right_invert: "(ba', t # bc) \<in> seq_lift R" 


text {* For each concrete behaviour there is a related abstract one. *}

lemma behaviour_refinement:
  "\<lbrakk> PO_refines R Ta Tc; bc \<in> beh Tc\<rbrakk> 
  \<Longrightarrow> \<exists>ba \<in> beh Ta. (ba, bc) \<in> seq_lift R" 
  apply (erule beh.induct, auto)
  -- {* case: singleton *}
  apply (clarsimp simp add: PO_refines_def Image_def)
  apply (drule subsetD, auto)
  -- {* case: cons; first construct related abstract state *}
  apply (erule sl_cons_right_invert, clarsimp)
  apply (rename_tac s bc s' ba t)
  -- {* now construct abstract transition *}
  apply (auto simp add: PO_refines_def PO_rhoare_def)
  apply (thin_tac "?X \<subseteq> ?Y")
  apply (drule subsetD, auto)
  done

definition 
  obs_consistent :: 
    "[('s \<times> 't) set, 'p \<Rightarrow> 'o, ('s, 'o) spec, ('t, 'p) spec] \<Rightarrow> bool"
where
  "obs_consistent R pi Sa Sc \<equiv> (\<forall>s t. (s, t) \<in> R \<longrightarrow> pi (obs Sc t) = obs Sa s)"

lemma obs_consistent_refl [iff]: "obs_consistent Id id S S"
by (simp add: obs_consistent_def)

lemma obs_consistent_trans [intro]: 
  "\<lbrakk> obs_consistent R1 pi1 S1 S2; obs_consistent R2 pi2 S2 S3 \<rbrakk>
  \<Longrightarrow> obs_consistent (R1 O R2) (pi1 o pi2) S1 S3"
by (auto simp add: obs_consistent_def)

lemma obs_consistent_empty: "obs_consistent {} pi Sa Sc"
by (auto simp add: obs_consistent_def)

lemma obs_consistent_conj1 [intro]: 
  "obs_consistent R pi Sa Sc \<Longrightarrow> obs_consistent (R \<inter> R') pi Sa Sc"
by (auto simp add: obs_consistent_def)

lemma obs_consistent_conj2 [intro]: 
  "obs_consistent R pi Sa Sc \<Longrightarrow> obs_consistent (R' \<inter> R) pi Sa Sc"
by (auto simp add: obs_consistent_def)

lemma obs_consistent_behaviours:
  "\<lbrakk> obs_consistent R pi Sa Sc; bc \<in> beh Sc; ba \<in> beh Sa; (ba, bc) \<in> seq_lift R\<rbrakk>
  \<Longrightarrow> map pi (map (obs Sc) bc) = map (obs Sa) ba"
by (erule seq_lift.induct) (auto simp add: obs_consistent_def) 

text {* Definition of refinement proof obligations. *}

definition 
  refines :: 
    "[('s \<times> 't) set, 'p \<Rightarrow> 'o, ('s, 'o) spec, ('t, 'p) spec] \<Rightarrow> bool" 
where
  "refines R pi Sa Sc \<equiv> obs_consistent R pi Sa Sc \<and> PO_refines R Sa Sc"

lemmas refines_defs = 
  refines_def PO_refines_def

lemma refinesI: 
  "\<lbrakk> PO_refines R Sa Sc; obs_consistent R pi Sa Sc \<rbrakk>
  \<Longrightarrow> refines R pi Sa Sc"
  by (simp add: refines_def)

lemma PO_refines_from_refines: 
  "refines R pi Sa Sc \<Longrightarrow> PO_refines R Sa Sc"
  by (simp add: refines_def)

lemma refinement_reflexive: "refines Id id S S"
  by (auto simp add: refines_defs)

lemma refinement_transitive: 
  "\<lbrakk> refines R1 pi1 S1 S2; refines R2 pi2 S2 S3 \<rbrakk> 
  \<Longrightarrow> refines (R1 O R2) (pi1 o pi2) S1 S3"
  apply (auto simp add: refines_defs del: subsetI
    intro: rhoare_trans)
  apply (fast dest: Image_mono)
  done

text {* Soundness of refinement for proving implementation *}

lemma observable_behaviour_refinement:
  "\<lbrakk> refines R pi Sa Sc; bc \<in> obeh Sc \<rbrakk> \<Longrightarrow> map pi bc \<in> obeh Sa"
  by (auto simp add: refines_def obeh_def image_def
         dest!: behaviour_refinement obs_consistent_behaviours)

theorem refinement_soundness: 
  "refines R pi Sa Sc \<Longrightarrow> implements pi Sa Sc"
  by (auto simp add: implements_def 
         elim!: observable_behaviour_refinement)

end
