(*<*)
theory LLInterface
imports ConcreteDefs
begin

(*>*)
section {* Low-level PMM interface definitions *}

text {*
  The following condition ensures that corruption will always
  be detectable; given a "wrong" piece contents (for the given
  instance), the condition ensures they will not matche the
  hashes.

*}

definition doesnt_match_hashes where
  "doesnt_match_hashes sp ind inst cnt \<equiv>
        (ind = Admin HASH_REG \<longrightarrow> \<not> ok_hash_cnt cnt)
        \<and> (ind \<noteq> Admin HASH_REG \<longrightarrow> 
            \<not> matches_htable (written_hash sp Primary) ind inst cnt
            \<and> \<not> matches_htable (written_hash sp Secondary) ind inst cnt
        )
  "

lemmas wrong_hashes_defs =
  doesnt_match_hashes_def
  ok_hash_cnt_def
  matches_htable_def
  written_hash_def

text {*

  This file describes the assumed behavior of the low-level read/write
  operations. Both are described through a non-deterministic choice of
  outcomes, with each outcome corresponding either to an expected
  result or to a particular kind of failure.

  On a technical note, all the operations here affect the persistent
  part of the state, which is represented by the following record
  type.

  \Snippet{state_def}

  The record is parametrized by the type of the contents which are
  written to the persistent state (type @{typ 'a}). The field @{term
  "memory"} maps the given instance (described as an index/instance
  combination) to its current content. The field @{term "hw_state"}
  describes the hardware state of each instance. The state can be
  either @{term "IOK"}, when the instance is operating normally; 
  @{term "IUnwr"}, when reliable writes are not possible any more,
  or @{term "IDead"}, when no reads (and by extension, no writes)
  are possible.

  We assume the absence of intermittent/transient failures which could
  arise from e.g. bus errors; in practical terms, this means that all
  these operations perform enough repeats to be confident in their
  result before failing and/or returning a value.

  Further, we assume that no "revival" of memory occurs. For instance,
  if the memory read of an area fails (the area is declared "dead"),
  we assume that all further attempts to read the same area will fail
  too; the same holds for writes. In other words, there is a linear
  ordering between the different hardware states, and none of the
  operations "upgrade" the state.  
*}

subsection {* Write operations *}

text {* 

  First, we describe a succesful write. The new content of the given
  instance is taken from the argument. 
*}

definition write_success 
  :: "indx \<Rightarrow> INST \<Rightarrow> 'c :: contents_cl \<Rightarrow> (unit, 'sv, 'c)pmm_monad" where
  "write_success ind inst cnt \<equiv> \<lambda>(sv, sp).
    {(Some (Inr ()), sv, sp')|sp'.
        size cnt \<le> capacity ind
        \<and> hw_state sp ind inst = IOK
        \<and> sp' = sp\<lparr> memory := fun_upd_second (memory sp) ind inst cnt \<rparr>
    }
  "

text {* 
  
  Technically, the definition can be read as follows: the @{term
  "write_success"} operation takes three arguments, @{term "ind"},
  @{term "inst"}, and @{term "cnt"}, of types @{typ "indx"}, @{typ
  "INST"}, and @{typ "'c"} respectively. The result is a function (a
  @{text "\<lambda>"}-expression) which takes a pair of values representing
  the volatile and persistent components of the current state, and
  returns the set of possible results. If the instance has been
  declared unwritable, this set is empty (i.e.  the operation is
  impossible). Otherwise, the result is a singleton set describing a
  successful write. In that case, the operation terminates normally,
  returning a "void" value @{term "()"}, and updates the persistent
  part of the state.

  The normal termination can be seen from @{term "Some (Inr ())"} as
  follows: @{term "Inr"} denotes a normal return value, as oposed to
  exceptional return values, which are denoted with @{term "Inl"}.
  Additionally, the operation does not restart, hence the @{term
  "Some"}. Restarting operations return the value @{term
  "None"}. Furthermore, the operation affects the persistent part of
  the state by updating its @{term "memory"} field.
  
 *}

text {*

  Next, we describe a write which fails due to a (permanent) memory
  error. 
*}

definition write_fail_instance 
  :: "indx \<Rightarrow> INST \<Rightarrow> 'c :: contents_cl \<Rightarrow> ('a, 'sv, 'c)pmm_monad" where
  "write_fail_instance ind inst cnt \<equiv> \<lambda>(sv, sp).
    {(Some (Inl FailWrite), sv, sp')|sp' cnt'. 
        size cnt \<le> capacity ind
        \<and> cnt' \<noteq> cnt
        \<and> (ind = Admin PTR_REG \<longrightarrow> cnt' \<in> range serialize_ptr)
        \<and> cnt' \<noteq> memory sp ind inst
        \<and> size cnt' \<le> capacity ind
        \<and> doesnt_match_hashes sp ind inst cnt'
        \<and> sp' = sp\<lparr> 
            memory := fun_upd_second (memory sp) ind inst cnt', 
            hw_state := fun_upd_second (hw_state sp) ind inst 
                (min (hw_state sp ind inst) IUnwr)
        \<rparr>
    }"

text {*

  The new instance content is an (almost) arbitrary value. Here we add
  a couple of assumptions. First, the error can only be detected if
  the resulting write differs from both the intended write and the
  previous contents of the memory (the case where it matches the
  previous contents will be dealt with later).

  Second, we assume that the failure will be detected with our
  checksums (hashes). This means that the new value (assumed to be
  random) cannot be successfully deserialized into a table containing
  hashes (such as the one stored in the hash region), and also that
  its hash does match the value stored for this instance in one of the
  hash region instances. These assumptions are reflected in the
  condition @{term "\<not>ok_hash cnt'"} and the one following it.  Note
  that, in reality, these assumptions can be violated, but only with a
  negligible probability.

  This operation is enabled even if the instance is unwritable and/or
  dead. The error returned is @{term "FailWrite"}. Additionally, the
  instance is (possibly redundantly) marked as "unwritable".

*}

text {*

  Another type of permanent memory failure is the one caused by the
  instance turning "read-only", i.e. the old content cannot be
  overwritten. This case is not covered by @{term
  "write_fail_instance"}, because the old content might actually
  successfully deserialize into a table of hashes or match the stored
  hash.
  
  We assume this kind of failure does not happen in the regions kept
  in the BBRAM (specifically the region containing hashes and the PTR
  region). Otherwise, this would change the behavior of our system.
  We could get "frozen" between two states, i.e. two versions of the
  contents. Whatever way we chose to break the tie, another hardware
  failure later could cause us to choose another version later, which
  is undesired, so the only safe option would be to halt the system.
  Furthermore, we would also need to fail in all states that might be
  descendants of such a "frozen" state, to prevent another undesired
  behavior of "reviving" the card after a failure.

*}

definition write_fail_readonly 
  :: "indx \<Rightarrow> INST \<Rightarrow> 'c :: contents_cl \<Rightarrow> ('a, 'sv, 'c)pmm_monad" where
  "write_fail_readonly ind inst cnt \<equiv> \<lambda>(sv, sp). 
    {(Some (Inl FailWrite), sv, sp')|sp'.
        size cnt \<le> capacity ind
        \<and> ind \<noteq> Admin HASH_REG
        \<and> ind \<noteq> Admin PTR_REG
        \<and> cnt \<noteq> memory sp ind inst
        \<and> sp' = sp\<lparr> hw_state := fun_upd_second (hw_state sp) ind inst
                            (min (hw_state sp ind inst) IUnwr)
        \<rparr>
  }"

text {*

  Next, we describe a write interrupted by a restart. The content
  written to the instance is again, almost arbitrary (same
  restrictions apply as for @{term "write_fail_instance"}). The
  difference is that here, the instance is not marked as "unwritable",
  because we might still be able to rewrite it with correct contents
  later. Finally, the volatile part of the memory is reset to an
  arbitrary value.

*}

definition write_corrupt_restart 
  :: "indx \<Rightarrow> INST \<Rightarrow> 'c :: contents_cl \<Rightarrow> ('a, 'sv \<times> 'c state)r_monad" where
  "write_corrupt_restart ind inst cnt \<equiv> \<lambda>(sv, sp). 
    {(None, sv', sp')|sv' sp' cnt'. 
        size cnt \<le> capacity ind
        \<and> cnt' \<noteq> cnt
        \<and> cnt' \<noteq> memory sp ind inst
        \<and> (ind = Admin PTR_REG \<longrightarrow> cnt' \<in> range serialize_ptr)
        \<and> size cnt' \<le> capacity ind
        \<and> doesnt_match_hashes sp ind inst cnt'
        \<and> sp' = sp\<lparr> 
            memory := fun_upd_second (memory sp) ind inst cnt' 
        \<rparr>
    }"

text {*
  
  Before performing any actual work, the write operation checks if
  input is oversized.

*}
definition write_too_large
  :: "indx \<Rightarrow> INST \<Rightarrow> 'c :: contents_cl \<Rightarrow> ('a, 'sv, 'c)pmm_monad" where
  "write_too_large ind inst cnt \<equiv> \<lambda>(sv, sp). 
    {(Some (Inl InvalidInput), sv, sp')|sp'. size cnt > capacity ind \<and> sp' = sp }"

text {* Finally, we describe the complete low-level write
operation. *}

definition write_f 
  :: "indx \<Rightarrow> INST \<Rightarrow> 'c :: contents_cl \<Rightarrow> (unit, 'sv, 'c)pmm_monad" where
  "write_f ind inst cnt \<equiv> doR {
    x \<leftarrow>  (write_success ind inst cnt 
    [+] write_fail_instance ind inst cnt
    [+] write_corrupt_restart ind inst cnt
    [+] write_fail_readonly ind inst cnt
    [+] write_too_large ind inst cnt
    [+] restart_mangle_sv);
    restart_mangle_sv [+] returnR x
}
  "

text {*

  We either perform one of the previously defined writes, or restart
  (resetting the volatile part of the state to an arbitrary
  value). After the operation has completed, we again allow the
  possibility of a restart (thus modeling e.g. the scenario where we
  restart right after a successful write).

  Technically, note that we use bindR here, instead of the regular
  bind, thus not distinguishing between exceptions and regular values

*}

(*<*)
lemmas write_f_defs = 
    write_f_def
    write_success_def
    write_fail_instance_def
    write_corrupt_restart_def
    write_fail_readonly_def
    write_too_large_def
    wrong_hashes_defs
(*>*)

subsection {* Read operations *}

text {*

  Read operations are similarly divided by their outcomes. We first
  describe a successful read. Note that this does not imply the
  "correctness" of the returned value in any way; we simply assume
  that the current memory contents are returned without any
  modification.

  The operation is only enabled if the instance has not been
  previously declared "dead". 

*}

definition ll_read_success 
  :: "indx \<Rightarrow> INST \<Rightarrow> ('c :: contents_cl, 'sv, 'c)pmm_monad" where
  "ll_read_success ind inst \<equiv> \<lambda>(sv, sp).
    {(Some (Inr (memory sp ind inst)), sv, sp')|sp'.
        hw_state sp ind inst \<noteq> IDead
        \<and> sp' = sp
    }"

text {*
  
  Next up, read operation modeling spontaneous corruption (e.g. due to bit
  flips). Note that, in reality, this corruption can take place long
  before the operation is executed; however, this is the first time we
  can observe it.

*}

definition ll_read_corrupt 
  :: "indx \<Rightarrow> INST \<Rightarrow> ('c :: contents_cl, 'sv, 'c)pmm_monad" where
  "ll_read_corrupt ind inst \<equiv> \<lambda>(sv, sp).
    {(Some (Inr cnt), sv, sp')|cnt sp'.
        hw_state sp ind inst \<noteq> IDead
        \<and> cnt \<noteq> memory sp ind inst
        \<and> (ind = Admin PTR_REG \<longrightarrow> cnt \<in> range serialize_ptr)
        \<and> size cnt \<le> capacity ind
        \<and> doesnt_match_hashes sp ind inst cnt
        \<and> sp' = sp\<lparr> 
            memory := fun_upd_second (memory sp) ind inst cnt 
        \<rparr>
    }
  "

text {*

  The instance contents are updated to an arbitrary value, which is
  then returned. Again, the restrictions on the resulting values are
  the same ones mentioned previously for @{term "write_fail_instance"}

  Note that no error is signaled at this level; the error
  detection/correction has to take place "higher up". The operation is
  only enabled if the instance has not been declared "dead".

*}

text {*

  Finally, we describe a failed read. The failure here is
  irreversible; we assume that no further operations on this instance
  are possible, neither reads nor writes. 

*}

definition ll_read_failure 
  :: "indx \<Rightarrow> INST \<Rightarrow> ('a, 'sv, 'c)pmm_monad" where
  "ll_read_failure ind inst \<equiv> \<lambda>(sv, sp).
    {(Some (Inl FailLLRead), 
        (sv, sp\<lparr> 
            hw_state := fun_upd_second (hw_state sp) ind inst IDead
        \<rparr>)
    )}
  "

text {*
  
  This is the complete low-level read operation. Same notes apply
  as for @{term "write_f"}.

*}

definition ll_read 
  :: "indx \<Rightarrow> INST \<Rightarrow> ('c :: contents_cl, 'sv, 'c)pmm_monad" where
  "ll_read ind inst \<equiv> doR{
    x \<leftarrow> (ll_read_success ind inst 
    [+] ll_read_failure ind inst 
    [+] ll_read_corrupt ind inst
    [+] restart_mangle_sv);
    restart_mangle_sv [+] returnR x
  }
  "

(*<*)
lemmas ll_read_defs = 
  ll_read_def 
  ll_read_success_def 
  ll_read_failure_def 
  ll_read_corrupt_def
  wrong_hashes_defs

end
(*>*)
