theory StartupDefs
  imports CommonDefs
begin

definition sync_hashes :: "INST \<Rightarrow> INST \<Rightarrow> (unit, pmm_vol_mem, 'c :: contents_cl)pmm_monad" where
  sync_hashes_def[monad_def]:
  "sync_hashes to from \<equiv> do {
    ht \<leftarrow> <try> get_hash_table from <catch> K (do { set_failed; throw FailMode });
    <try>
        write_f (Admin HASH_REG) to (htable_to_hinst_cnt ht)
    <catch> K set_degraded_h
  }"

text {* The first operation performed after startup: checking the hash
  region. After the operation, if both instances are correct, they
  should be the same.

  Otherwise, at least one of them has failed. However, if just one is
  failed, we do not want to bail out with a DegradedH exception here;
  it might be "too optimistic", in that there actually might exist a
  failed region. Thus we swallow the write exceptions and continue
  with the startup checks (setting the degraded_h flag)
*}

definition check_hash_region :: "(unit, pmm_vol_mem, 'c :: contents_cl)pmm_monad" where
  check_hash_region_def[monad_def]:
  "check_hash_region \<equiv> <try> do {
    ht2 \<leftarrow> get_hash_table Secondary;
    <try> do {
        ht1 \<leftarrow> get_hash_table Primary;
        (if ht1 \<noteq> ht2
            then 
                <try>
                    write_f (Admin HASH_REG) Primary (htable_to_hinst_cnt ht2) 
                <catch> (K set_degraded_h)
            else return ()
        )
    } <catch>
        (\<lambda>e. case e of 
            FailRead \<Rightarrow> 
                <try>
                    write_f (Admin HASH_REG) Primary (htable_to_hinst_cnt ht2) 
                <catch> (K set_degraded_h)
            | FailLLRead \<Rightarrow> set_degraded_h
            | e' \<Rightarrow> throw e'
        )
  }
  <catch>
    (\<lambda>e. case e of 
        FailRead \<Rightarrow> sync_hashes Secondary Primary
        | FailLLRead \<Rightarrow> do {
            (<try> get_hash_table Primary <catch> (K (do { set_failed; throw FailMode })));
            set_degraded_h
        }
        | _ \<Rightarrow> throw e
    )
  "

text {*
  This is the "general" version of the PTR, where we branch out
  depending on the status of the hash region. In case of a hash
  region degradation, we want to:
    1. avoid writing anything to the hash region
    2. still make sure that no two different, but individually correct
       versions of the PTR exist. Thus, if two clashing versions exist,
       we "mess up" one of them on purpose.

  Note that the @{term "check_ptr_DegradedH"} operation is declared
  in CommonDefs, since it is also used in UpdateDefs
*}

text {*

  If we're in DegradedH, we cannot sync the PTRs anymore, even though
  they might be out of sync. To prevent the situation where we are
  "frozen" between two correct different PTR values (e.g. one saying
  that we should essentially "sync right" and the other that we should
  "sync left"), we purposefully make one of them incorrect.
*}

text {*
  Next, we check the PTR. The requirements are analogous to the ones for
  @{term "check_hash_region"}. Similar to there, we swallow the Degraded
  exception here, so that we still sync the public regions.

  This version of check_ptr assumes that no degradation of hash regions
  has been detected.
*}

definition check_ptr :: "(unit, pmm_vol_mem, 'c :: contents_cl)pmm_monad" where
  check_ptr_def[monad_def]:
  "check_ptr \<equiv> <try> do {
    ptr_set2 \<leftarrow> read_ptr_instance Secondary;
    <try> do {
        ptr_set1 \<leftarrow> read_ptr_instance Primary;
        (if (ptr_set2 \<noteq> ptr_set1)
            then
                <try>
                    write_ptr_instance Primary ptr_set2
                <catch> (\<lambda>e. case e of
                    Degraded v \<Rightarrow>
                        if v = EB False
                        then
                            do {
                                set_degraded_p;
                                <try> 
                                    write_f (Admin PTR_REG) Primary (serialize_ptr ptr_set2)
                                <catch> K (return ())
                            }
                        else return ()
                    | _ \<Rightarrow> throw e
                )
            else return ())
    }
    <catch> 
        (\<lambda>e. case e of 
            FailRead \<Rightarrow>
                <try>
                    write_ptr_instance Primary ptr_set2
                <catch> (\<lambda>e. case e of
                    Degraded _ \<Rightarrow> set_degraded_p
                    | _ \<Rightarrow> throw e
                )
            | Degraded v \<Rightarrow>
                if v = EU
                then set_degraded_p
                else
                    if (parse_ptr (get_exc_val_cnt v) \<noteq> ptr_set2)
                    then <try> write_f (Admin PTR_REG) Primary (serialize_ptr ptr_set2) 
                         <catch> K (return ())
                    else return ()
            | _ \<Rightarrow> throw e
        )
  }
  <catch> (\<lambda>e. case e of 
    Degraded v \<Rightarrow>
        if v = EU
        then do { read_ptr_instance_or_fail_silence_deg Primary; set_degraded_p }
        else
            <try>
                write_f (Admin PTR_REG) Primary (serialize_ptr (parse_ptr (get_exc_val_cnt v)))
            <catch> K (return ())
    | FailRead \<Rightarrow>
        <try>
            just_sync_ptr Secondary Primary
        <catch> (\<lambda>e. case e of Degraded _ \<Rightarrow> return () | _ \<Rightarrow> throw e)
    | _ \<Rightarrow> throw e)
  "

text {*

  Finally, we perform the checks for the public regions. Requirements
  are similar to those for the PTR and hash region, but the checks are
  now more complex:

  1. it is no longer sufficient for an instance to match its written
     hash for it to be correct; we also need to make sure that it's a
     part of the last commited transaction. For this, we need to
     compare hashes of both instances, and the refer to the PTR in
     case they don't match

  2. the failure modes are also more complex: every check also needs
     to deal with a potential degradation of both the hash region and
     the PTR

  The first version performs the core check, assuming that the card is
  not in DegradedH mode.

*}

definition check_pub_same
    :: "pub_indx  \<Rightarrow> (unit, pmm_vol_mem, 'c :: contents_cl)pmm_monad" where
  check_pub_same_def[monad_def]:
  "check_pub_same pind \<equiv>
    <try> do {
        read (Public pind) Primary;
        (<try> do {
            read (Public pind) Secondary;
            return ()
        } <catch>
            (\<lambda>e. case e of
                FailRead \<Rightarrow> just_sync_silence_deg Secondary Primary pind
                | Degraded _ \<Rightarrow> return ()
                | _ \<Rightarrow> throw e
            )
        )
   } <catch> (\<lambda>e. case e of
        FailRead \<Rightarrow> just_sync_silence_deg Primary Secondary pind 
        | Degraded v \<Rightarrow>
            if v = EU
            then read_or_fail_silence_deg (Public pind) Secondary \<guillemotright> return ()
            else return ()
        | _ \<Rightarrow> throw e
    )
  "

definition check_pub_region
  :: "pub_indx \<Rightarrow> (unit, pmm_vol_mem, 'c :: contents_cl)pmm_monad" where
  check_pub_region_def[monad_def]:
  "check_pub_region pind \<equiv> do {
    dir \<leftarrow> get_sync_direction pind;
    if dir = SEq
    then do { check_pub_same pind; remove_from_ptr_silence_deg pind }
    else
        if dir = SR
        then sync_and_remove_from_ptr_silence_deg Secondary Primary pind 
        else sync_and_remove_from_ptr_silence_deg Primary Secondary pind 
  }"
 
definition check_all_public :: "
  (unit, pmm_vol_mem, 'c :: contents_cl) pmm_monad" where
  check_all_public_def[monad_def]:
  "check_all_public \<equiv> do {
    l \<leftarrow> chooseN (order_set UNIV);
    mapM' check_pub_region l
  }"

definition startup :: "('a, pmm_vol_mem, 'c :: contents_cl)pmm_monad" where
  startup_def[monad_def]:
  "startup \<equiv> do {
    init_vars;
    check_hash_region;
    check_ptr;
    check_all_public;
    throw_mode_error EU;
    throw StartedExc
  }"

end
