signature DEBUG_MATCH =
sig
  val first_order_match: Proof.context -> theory -> term * term
    -> Type.tyenv * Envir.tenv -> Type.tyenv * Envir.tenv
end

structure Debug_Match : DEBUG_MATCH =
struct

(*Tests whether 2 terms are alpha/eta-convertible and have same type.
  Note that Consts and Vars may have more than one type.*)
fun t aeconv u = t aconv u orelse
  Envir.eta_contract t aconv Envir.eta_contract u;

fun mapbnd f =
    let fun mpb d (Bound(i))     = if i < d then Bound(i) else Bound(f(i-d)+d)
          | mpb d (Abs(s,T,t))   = Abs(s,T,mpb(d+1) t)
          | mpb d ((u1 $ u2))    = (mpb d u1)$(mpb d u2)
          | mpb _ atom           = atom
    in mpb 0 end;

val incr = mapbnd (fn i => i+1);

(*** Matching ***)

fun deb_typ_match tsig =
  let
    fun match (V as TVar (v, S), T) subs =
          (case Type.lookup subs (v, S) of
            NONE =>
              if V = T then subs
              else if Type.of_sort tsig (T, S) then Vartab.update_new (v, (S, T)) subs
              else (
                  tracing "TVar/type mismatch; nothing in the vartabl";
                  pprint (pretty_typs @{context} [V, T]);
                  raise Type.TYPE_MATCH
                  )
          | SOME U => if U = T then subs else (
                      tracing "TVar/type mismatch; vartab entry found";
                      pprint (pretty_typs @{context} [V, T]);
                      tracing "vartab entry:";
                      pprint (pretty_typ @{context} U);
                      raise Type.TYPE_MATCH
                      )
          )
      | match (Type (a, Ts), Type (b, Us)) subs =
          if a <> b then (
              tracing "(constant) Type mismatch";
              pprint (pretty_typs @{context} [Type (a, Ts), Type (b, Us)]);
              raise Type.TYPE_MATCH
              )
          else matches (Ts, Us) subs
      | match (TFree x, TFree y) subs =
          if x = y then subs else (
              tracing "TFree mismatch";
              pprint (pretty_typs @{context} [TFree x, TFree y]);
              raise Type.TYPE_MATCH
          )
      | match (T, U) _ = (
            tracing "Different type constructors";
            pprint (pretty_typs @{context} [T, U]);
            raise Type.TYPE_MATCH
        )
    and matches (T :: Ts, U :: Us) subs = matches (Ts, Us) (match (T, U) subs)
      | matches _ subs = subs;
  in match end;
    

fun typ_match thy (TU as (T, U)) tyenv = Sign.typ_match thy TU tyenv
  handle Type.TYPE_MATCH => (
         tracing "type mismatch";
         pprint (pretty_typs @{context} [T, U]);
         tracing "type environment:";
         pprint (pretty_tyenv @{context} tyenv);
         ((deb_typ_match (Sign.tsig_of thy) TU tyenv) handle Type.TYPE_MATCH => raise Pattern.MATCH);
         raise Pattern.MATCH
         );

(*First-order matching;
  The pattern and object may have variables in common.
  Instantiation does not affect the object, so matching ?a with ?a+1 works.
  Object is eta-contracted on the fly (by eta-expanding the pattern).
  Precondition: the pattern is already eta-contracted!
  Types are matched on the fly*)
fun first_order_match ctxt thy =
  let
    fun mtch k (instsp as (tyinsts,insts)) = fn
        (Var(ixn,T), t)  =>
          if k > 0 andalso Term.is_open t then raise Pattern.MATCH
          else (case Envir.lookup' (insts, (ixn, T)) of
                  NONE => (typ_match thy (T, fastype_of t) tyinsts,
                           Vartab.update_new (ixn, (T, t)) insts)
                | SOME u => if t aeconv u then instsp else
                            (tracing "Mismatch:";
                             pprint (pretty_terms ctxt [t, u]);
                             raise Pattern.MATCH))
      | (t as Free (a,T), u as Free (b,U)) =>
          if a=b then (typ_match thy (T,U) tyinsts, insts) else
            (
                tracing "Free mismatch:";
                pprint (pretty_terms ctxt [t, u]);
                raise Pattern.MATCH
            )
      | (t as Const (a,T), u as Const (b,U))  =>
          if a=b then (typ_match thy (T,U) tyinsts, insts) else
          (
                tracing "Free mismatch:";
                pprint (pretty_terms ctxt [t, u]);
                raise Pattern.MATCH
          )
      | (Bound i, Bound j)  =>
        if  i=j  then  instsp
        else ( tracing "Bound mismatch"; raise Pattern.MATCH)
      | (Abs(_,T,t), Abs(_,U,u))  =>
          mtch (k + 1) (typ_match thy (T,U) tyinsts, insts) (t,u)
      | (f$t, g$u) => mtch k (mtch k instsp (f,g)) (t, u)
      | (t, Abs(_,_,u))  =>  mtch (k + 1) instsp ((incr t)$(Bound 0), u)
      | (t, u) =>
        (tracing "Mismatched kinds of terms";
         pprint (pretty_terms ctxt [t, u]);
         raise Pattern.MATCH
        )
  in fn tu => fn env => mtch 0 env tu end;

end
