(*
  Simproc for a Java-like "checked exceptions" reasoning. 
*)
signature NO_THROW = sig
    val mk_goal_try_catch: Proof.context -> term -> term -> term
    val mk_goal_bindN: Proof.context -> term -> term -> term
    val throw_simproc_simple': thm list -> (Proof.context -> term -> term -> term) -> string -> simpset -> term -> thm option
    val throw_simproc': string -> simpset -> term -> thm option
    val no_throw_debug: bool Config.T
end

structure No_Throw : NO_THROW =
struct

structure No_Throw_Thms = Named_Thms
    (val name = @{binding "no_throw"}
     val description = "Theorems for pruning paths by checking the exceptions thrown")

val _ = (Context.>> o Context.map_theory) No_Throw_Thms.setup

val (no_throw_debug, no_throw_debug_setup) = 
    Attrib.config_bool @{binding "no_throw_debug"} (K false)
val _ = (Context.>> o Context.map_theory) no_throw_debug_setup

fun is_exceptions_case t = is_comb t andalso 
                           const_name (get_fun_term t) = @{const_name "exceptions_case"}

fun is_funtype (Type ("fun", [_, _])) = true
  | is_funtype _ = false;

fun mk_goal_gen d_throw ctxt t e = let
    (*
    val _ = tracing "Making goal:"
    val _ = pprint (pretty_terms ctxt [d_throw, t, e])
            *)

    (*
    val _ = tracing "t type"
    val _ = pprint (pretty_typ ctxt (type_of t))
    val _ = tracing "e type"
    val _ = pprint (pretty_typ ctxt (type_of e))
     *)

    val t' = apply_typed ctxt d_throw t

    fun process_fun _ = let
        (*
        val _ = tracing "mk_goal_gen/process_fun"
         *)
        val T = domain_type (type_of e)
                
        (*
        val _ = tracing "range type:"
        val _ = pprint (pretty_typ ctxt T)
         *)

        val fvar = fold Term.variant_frees [t, e] [("x", T)] |> hd
        (*
        val _ = tracing "before first apply_typed'"
        val e' = apply_typed' ctxt e (Free fvar)
        val _ = tracing "after first apply_typed'"
        val _ = tracing "fvar:"
        val _ = tracing (PolyML.makestring (Free fvar))
         *)
        val e' = e $ Free fvar
        val (fvarN, fvarT) = fvar
    in
        apply_typed' ctxt t' e'
                     |> (fn (hyp, subst) => HOLogic.mk_all (fvarN, subst fvarT, hyp))
    end

    val hyp = if (is_funtype (type_of e))
              then process_fun ()
              else (
                  (*
                  tracing "mk_goal_gen: not a funtype";
                   *)
                  apply_typed' ctxt t' e
                               |> fst
                  )
in
    hyp
        |> HOLogic.mk_Trueprop
end

val mk_goal_try_catch = mk_goal_gen @{term_pat "doesnt_throw"}
                                 
val mk_goal_bindN = mk_goal_gen @{term_pat "doesnt_throw_raw"}
                            

(* Handles the simple case of either: 
                              
   @{term "m \<guillemotright>=N exceptions_case h1 h2 h3 h4 h5 h6 h7 h9"}
   or
   @{term "try_catch m exceptions_case h1 h2 h3 h4 h5 h6 h7 h9"}
   or
   @{term "m \<guillemotright>=N exceptions_case (h1 e) (h2 e) (h3 e) (h4 e) (h5 e)
                                 (h6 e) (h7 e) (h9 e)"};
 *)

fun throw_simproc_simple' thms mk_goal tyname ss redex =
    let
        val ctxt = Simplifier.the_context ss
        val thy = Proof_Context.theory_of ctxt

        val trace = Config.get ctxt no_throw_debug

        fun cond_trace x = if trace then (tracing x) else ()
        fun cond_pprint x = if trace then (pprint x) else ()
        val _ = cond_trace "throw_simproc_simple': redex is"
        val _ = cond_pprint (pretty_term ctxt redex)

        val (lhs, rhs) = dest_binop redex

        val exc_constrs = Datatype_Data.get_constrs thy tyname
                                                    |> the
                                                    |> map Term.Const
                                                    (*
                                                    |> map Logic.varify_global
                                                     *)

        fun get_case t i = 
            let 
                val total_cases = length exc_constrs
                (* val _ = tracing (PolyML.makestring t) *)
                fun get_app 0 (_ $ r) = r
                  |   get_app i (l $ _) = get_app (i-1) l
            in
                get_app (total_cases - i) t
            end

        val cases = let
            (* val exc_case_term = Envir.eta_contract rhs *)
            val exc_case_term =
                rhs
                    |> Envir.eta_contract
                    |> mk_abs_free
                    |> snd
                    |> strip_comb
                    |> snd
        in
            take (length exc_constrs) exc_case_term 
            (* map (get_case exc_case_term) (1 upto (length exc_constrs)) *)
        end

        val _ = (cond_trace "Cases:"; cond_pprint (pretty_terms ctxt cases))

        fun not_empty t =
            not (Pattern.matches thy (@{term_pat empty}, strip_abs_body t))

        (*
        val _ = (cond_trace "Theorems:"; cond_pprint (pretty_thms ctxt (No_Throw_Thms.get ctxt)))
         *)

        fun try_solve goal = 
            let 
                val mod_ss = ss addsimps (No_Throw_Thms.get ctxt)
                val stac = (asm_full_simp_tac mod_ss 1)
            (*
                val _ = cond_trace "Solving:";
                val _ = cond_pprint (pretty_term ctxt goal)
             *)
            in
                Goal.prove ctxt [] [] goal
                           (K (CHANGED stac))
                           |> SOME
                handle ERROR t => (cond_trace ("Prove error:" ^ t); NONE)
            end

        val thms_excs = 
            thms ~~ exc_constrs ~~ cases
                 |> filter (not_empty o snd)
                 |> map (fst ##> try_solve o mk_goal ctxt lhs)
                 |> tap (map (fn (_, x) => (
                                 (if is_some x
                                  then (cond_trace "Success;"; cond_pprint (pretty_thm ctxt (the x)))
                                  else cond_trace "Failed")
                        )))
                 |> filter (is_some o snd)
                 |> map (fn (x, y) => (x, the y))
                 |> map (fn (x, y) => y RS x)
            
    in
        if null thms_excs
        then NONE
        else SOME (
             (foldr1 (fn (x, y) => x RS (y RS @{thm trans})) thms_excs)
                 RS @{thm "eq_reflection"}
             ) 
    end

fun is_exceptions_case_or_abs t = 
    is_exceptions_case (strip_abs_body t)

fun throw_simproc' tyname ss redex = let
    val (_, rhs) = dest_binop redex
    val ctxt = Simplifier.the_context ss
    val thy = Proof_Context.theory_of ctxt
                  
    val complex_pat =
        Logic.varify_global
        @{term "(m :: 'a \<Rightarrow> ('b exceptions \<times> 'c) set)
        \<guillemotright>=N
        (\<lambda>e. (exceptions_case
        ((m1 :: 'b exceptions \<Rightarrow> 'c \<Rightarrow> ('d \<times> 'e) set) e)
        (m2 e) (m3 e) (m4 e) (m7 e) (m8 e) (m9 e)) e
        \<guillemotright>=N n)"}
    
    val target_pat =
        Logic.varify_global
        @{term "(m :: 'a \<Rightarrow> ('b exceptions \<times> 'c) set)
        \<guillemotright>=N
        (\<lambda>e. (exceptions_case
        ((m1 :: 'b exceptions \<Rightarrow> 'c \<Rightarrow> ('d \<times> 'e) set) e)
        (m2 e) (m3 e) (m4 e) (m7 e) (m8 e) (m9 e)) e)"} 
    
    val reassoc_thm = @{thm "bindN_assoc"} RS @{thm "sym"} 
    fun reassoc tm = let
        val init_env = (Vartab.empty, Vartab.empty)
        val (_, fo_env) = Pattern.match thy (complex_pat, tm) init_env
    in
        Envir.subst_term (Vartab.empty, fo_env) target_pat
    end
                     
    val trace = Config.get ctxt no_throw_debug
    fun cond_trace t = if trace then (tracing t) else ()
    fun cond_pprint t = if trace then (pprint t) else ()

    val _ =  (if is_exceptions_case_or_abs rhs
              then (cond_trace "No rewriting")
              else (
                  cond_trace "Needs rewriting"
                  ); cond_trace "Full redex: "; cond_pprint (pretty_term ctxt redex ))

    val dtr_thms = @{thms doesnt_throw_raw_bindN}

    fun maybe_thm tm = 
        tm 
            |> tap (fn x => 
                       (cond_trace "throw_simproc': rewriting: "; cond_pprint (pretty_term ctxt x)
                       )
                   )
            |> reassoc
            |> tap (fn x => 
                       (cond_trace "after reassoc"; cond_pprint (pretty_term ctxt x)) )
            |> Envir.beta_eta_contract
            |> tap (fn x => 
                       (cond_trace "after beta_eta"; cond_pprint (pretty_term ctxt x)) )
            (*
             |> unbindN
                |> fst
                |> tap (fn x => 
                           (cond_trace "after unbindN/fst"; cond_pprint (pretty_term ctxt x)) )
             (* TODO: figure out why this fails *)
             *)
            |> (fn t => (throw_simproc_simple' dtr_thms mk_goal_bindN tyname ss t) handle _ => NONE)
    (* |> tap (fn x => if is_some x then cond_pprint (pretty_thm ctxt (the x)) else ()) *)

        
    fun make_thm () =
        maybe_thm redex 
                  bind_opt
                  (fn x => SOME (
                           x
                               |> (fn x => x RS @{thm meta_eq_to_obj_eq})
                               |> (fn x => x RS @{thm bindN_lcong})
                               |> (fn x => x RS (reassoc_thm RS @{thm trans}))
                               |> (fn x => x RS @{thm eq_reflection})
                  (*
                   |> tap (fn x => if trace 
                                   then cond_pprint (pretty_thm ctxt x)
                                   else ()
                          )
                   *)
                  ))

in
    (if is_exceptions_case_or_abs rhs
     then (throw_simproc_simple' dtr_thms mk_goal_bindN tyname ss redex) 
     else make_thm())
        bind_opt (fn th => ((if trace 
                             then (
                                 cond_trace "throw simproc': final result";
                                 cond_pprint (pretty_thm ctxt th);
                                 cond_trace "throw simproc': end of result"
                                 ) else ()); SOME th))
end


end
