signature PMM_TACS = sig
    val pmm_branch_match_tac : Proof.context -> thm list -> thm list -> int -> tactic
    val clarsimp_adj_tac: string * (simpset -> int -> tactic) -> Proof.context -> int -> tactic
    val ex_eq_simproc: simproc
    val ex_eq_simproc_setup: Context.generic -> Context.generic
    val full_dest_tac: Proof.context -> thm list -> tactic
end

structure Pmm_Tacs: PMM_TACS = 
struct

val rmonad_splits = [
        @{thm "option.split"} 
        , @{thm "sum.split"} 
        , @{thm "split_if"} 
        , @{thm "exceptions.split"}
        , @{thm "option.split_asm"} 
        , @{thm "sum.split_asm"} 
        , @{thm "split_if_asm"} 
        , @{thm "exceptions.split_asm"}
]


fun pmm_branch_match_tac' ctxt ss = 
let
  val add_rmonad_splits = fold Splitter.add_split rmonad_splits
  fun add_ss_to_ctxt ss = map_simpset (fn s => add_rmonad_splits (s addsimps ss))
  (* fun my_auto ctxt = (auto_tac (add_ss_to_ctxt @{thms memory_upd_simps} ctxt )) *)
  fun my_auto ctxt = (auto_tac (add_ss_to_ctxt (@{thm fun_upd_twist}::ss) ctxt ))
in
      clarsimp_tac (add_ss_to_ctxt ss ctxt)
          THEN_ALL_NEW
          (
              (SELECT_GOAL (IF_UNSOLVED
                  (my_auto ctxt)
              ))
          )
end

fun pmm_branch_match_tac ctxt ss_prem ss_rest = 
let
  val full_ss = ss_prem @ ss_rest

  fun aux_tac i =
  ((SOLVED'
      (
          (rtac @{thm choice_subsetI1} THEN' pmm_branch_match_tac' ctxt full_ss)
          ORELSE' (pmm_branch_match_tac' ctxt full_ss)
      )
  ) ORELSE'
  (
      SOLVED' (rtac @{thm choice_subsetI2} THEN' (fn i => aux_tac i))
  )) i
in
  aux_tac
end

(*

 Version of clarsimp which allows one to choose the simplifier
 tactic (simp_tac, asm_lr_simp_tac etc)

 Adapted from clasimp.ML

*)

fun clasimp f name tac ctxt =
  Classical.map_claset (fn cs => f (cs, (name, CHANGED o tac (simpset_of ctxt)))) ctxt;

fun addSss (desc, tac) = clasimp Classical.addSafter desc tac;

val no_asmN = "no_asm";
val no_asm_useN = "no_asm_use";
val no_asm_simpN = "no_asm_simp";
val asm_lrN = "asm_lr";
 
val simp_options =
 (Args.parens (Args.$$$ no_asmN) >> K ("simp (no_asm)", simp_tac) ||
  Args.parens (Args.$$$ no_asm_simpN) >> K ("simp (no_asm_simp)", asm_simp_tac) ||
  Args.parens (Args.$$$ no_asm_useN) >> K ("simp (no_asm_use)",  full_simp_tac) ||
  Args.parens (Args.$$$ asm_lrN) >> K ("simp (asm_lr)", asm_lr_simp_tac) ||
  Scan.succeed ("simp", asm_full_simp_tac));

fun clarsimp_adj_tac tac ctxt =
  (snd tac) (simpset_of ctxt) THEN_ALL_NEW
  Classical.clarify_tac (addSss tac ctxt);

val clarsimp_adj_method =
  Scan.lift simp_options --|
    Method.sections (Clasimp.clasimp_modifiers) >>
    (fn tac => fn ctxt => SIMPLE_METHOD' (CHANGED_PROP o clarsimp_adj_tac tac ctxt));

val clarsimp_adj_all_method =
  Scan.lift simp_options --|
    Method.sections (Clasimp.clasimp_modifiers) >>
    (fn tac => fn ctxt => SIMPLE_METHOD 
        ((CHANGED_PROP o PARALLEL_GOALS o ALLGOALS) (clarsimp_adj_tac tac ctxt)))

val clarsimp_adj_setup = Method.setup @{binding clarsimp_adj} clarsimp_adj_method
  "clarify simplified goal, adjustable simplification";

val clarsimp_adj_all_setup = Method.setup @{binding clarsimp_adj_all} clarsimp_adj_all_method
  "clarify simplified goal, all goals, adjustable simplification";

(*

  Simproc for transforming terms of form @{term "\<exists>x. P x \<and> Q x"}
  where @{term "Q x"} contains a term of form @{term "x = a"} into terms of
  form @{term "\<exists>x. x = a \<and> P x \<and> Q' x"}, which can then be
  simplified into @{term "P a \<and> Q' a"} by HOL.simp_thms

*)

fun is_and (Const ("HOL.conj", _) $ _ $ _) = true
  | is_and _ = false

fun is_ex (Const ("HOL.Ex", _) $ Abs _) = true
  | is_ex _ = false

fun strip_ex (Const ("HOL.Ex", _) $ Abs (var, ty, t)) = ([Free (var, ty)], t)
  | strip_ex t = ([], t)

val left_conj = Pmm_Library.get_arg 1 
val right_conj = Pmm_Library.get_arg 0 

fun is_assgn_of t t' = case t' of
                           (Const ("HOL.eq", _) $ tl $ tr) => t = tl orelse t = tr
                         | _  => false

fun repeat1_conv cv = cv then_conv Conv.repeat_conv cv

fun find_eq ctxt bounds t = let
  val pull_ex_out = @{thm pull_ex_out} RS eq_reflection
  val conj_commute = @{thm conj_commute} RS eq_reflection
  val conj_assoc = @{thm conj_assoc} RS eq_reflection
  val ex_comm = @{thm ex_comm} RS eq_reflection

  val assoc_conv = Conv.rewr_conv conj_assoc
  val comm_conv = Conv.rewr_conv conj_commute

  (* Have to write this eta-abstracted to prevent infinite recursion *)
  fun pull_conv ctxt ct =  (
        (Conv.rewr_conv pull_ex_out
            then_conv Conv.try_conv (Conv.binder_conv
                (K (Conv.binder_conv (K assoc_conv) ctxt))
                ctxt
            )
        )
        else_conv
            Conv.binder_conv (K (pull_conv ctxt)) ctxt
        ) ct

  (* Perform the conversion through as many layers of binders as possible *)
  fun push_conv ctxt ct = (
        Conv.rewr_conv ex_comm
        then_conv (Conv.try_conv (Conv.binder_conv (K (push_conv ctxt)) ctxt))
    ) ct

  fun proc_and t = let
      val (left, right) = (left_conj t, right_conj t)
      val try_left = find_eq ctxt bounds left 
  in 
      if is_some try_left
        then map_opt Conv.arg1_conv try_left
        else (let val try_right = find_eq ctxt bounds right in
            map_opt (fn cv =>
                Conv.arg_conv cv then_conv comm_conv then_conv
                        (Conv.try_conv assoc_conv)) try_right
            end)
  end

  fun proc_ex t = let
      val ([f_var], body) = strip_ex t
      val body' = subst_bound (f_var, body)
      val sub_conv = find_eq ctxt (bounds @ [f_var]) body'
  in
    map_opt (fn cv =>
        Conv.binder_conv (K cv) ctxt then_conv Conv.try_conv (
            repeat1_conv (pull_conv ctxt)
            else_conv (push_conv ctxt)
        )
    ) sub_conv
  end

in
    if (bounds <> [] andalso is_assgn_of (hd bounds) t) 
        then SOME Conv.all_conv
        else if is_and t
            then proc_and t
            else if is_ex t
                then proc_ex t
                else NONE
end

fun ex_eq_do_proc ss t =
let
  val ctxt = Simplifier.the_context ss
  val thy = Proof_Context.theory_of ctxt
  val pull_out_conv = find_eq ctxt [] t
  val rewr_thms = map (fn x => x RS eq_reflection) [@{thm simp_thms(39)}, @{thm simp_thms(40)}]

  (* Throws a CTERM on terms of form  @{term "\<exists>x. x = t"} *)
  fun apply_eq_conv ctxt ct = (
      Conv.rewrs_conv rewr_thms
      else_conv Conv.binder_conv (K (apply_eq_conv ctxt)) ctxt
  ) ct

  val res_conv = map_opt (fn cv => cv then_conv Conv.try_conv (apply_eq_conv ctxt)) pull_out_conv 

in
    map_opt (fn cv => cv (cterm_of thy t)) res_conv
end
 
val ex_eq_simproc =
let
  val thy = @{theory}

  val pat = [
      @{term "\<exists>x. P x"}
  ]

in 
  Simplifier.simproc_global_i thy "ex_eq_simproc" pat (K ex_eq_do_proc)
end

val ex_eq_simproc_setup =
  Simplifier.map_ss (fn ss => ss addsimprocs [ex_eq_simproc]);


val _ = (Context.>> o Context.map_theory) clarsimp_adj_setup

val _ = (Context.>> o Context.map_theory) clarsimp_adj_all_setup

        (*
val _ = (Context.>> o Context.map_ss) ex_eq_simproc_setup

val _ = Addsimprocs [ex_eq_simproc_setup] 
        *)

fun cartesian (l::ls) = map (fn x => map (cons x) (cartesian ls)) l |> flat
  | cartesian [] = [[]]
                       
(* Check whether we can unify all the terms in ts to
some terms in ts' simultaneously *)

fun try_unify_all ctxt ts (thms : thm list) = let
    val thy = Proof_Context.theory_of ctxt
    fun matching_thms t = filter (fn th => Term.could_unify (t,prop_of th)) thms
    val matches_ts = map matching_thms ts
    fun same_thms (th, th') = (prop_of th) aconv (prop_of th')
    val possibilities = filter (not o has_duplicates (same_thms)) (cartesian matches_ts)
in
    find_first (fn ths => Unify.matches_list thy ts (map prop_of ths)) possibilities
end

fun all_prems_satisfied ctxt thm = try_unify_all ctxt (prems_of thm)
                     
fun mk_full_dest_tac thms ({context, prems,...} : Subgoal.focus) = let
    val prem_solvers =
        map (fn th => (th, all_prems_satisfied context th prems)) thms
            |> filter (is_some o snd)
            |> map (apsnd the)
    val satisfiable = map (fn (th, prems) => th OF prems) prem_solvers
in
    Method.insert_tac satisfiable 1
end

fun full_dest_tac ctxt thms =
    ALLGOALS (Subgoal.FOCUS (mk_full_dest_tac thms) ctxt)

end
