signature REF_THMS =
sig
  val get: Proof.context -> thm list
  val get_full: Proof.context -> (term * thm list) list
  val get_term_ref: Proof.context -> term -> thm list
  val add_thm: thm -> Context.generic -> Context.generic
  val add_thms: thm list -> Context.generic -> Context.generic
  val del_thm: thm -> Context.generic -> Context.generic
  val extract_sim_targets: term -> term * term
  val extract_sim_targets_thm: thm -> term * term
  val transfer: Proof.context -> Context.generic -> Context.generic
  val add: attribute
  val del: attribute
  val get_bind_thms: Proof.context -> thm list
  val get_try_catch_thms: Proof.context -> thm list
  val get_exc_case_thms: Proof.context -> thm list
  val get_if_thms: Proof.context -> thm list
  val get_if_full_thms: Proof.context -> thm list
  val get_choice_thms: Proof.context -> thm list
  val get_mapM_thms: Proof.context -> thm list
  val get_bind_bindS_thms: Proof.context -> thm list
  val get_try_catch_bindS_thms: Proof.context -> thm list
  val get_mapM_bindS_thms: Proof.context -> thm list
  val get_bindS_thms: Proof.context -> thm list
  val get_sim_defs: Proof.context -> thm list
  val get_simp_thms: Proof.context -> thm list
  val get_cong_thms: Proof.context -> thm list
end

structure Ref_Thms : REF_THMS =
struct

structure Ref_Bind_Thms = Named_Thms
    (val name = @{binding "ref_bind"}
     val description = "Theorems for automatic refinement over bind")

structure Ref_Try_Catch_Thms = Named_Thms
    (val name = @{binding "ref_try_catch"}
     val description = "Theorems for automatic refinement over try_catch")

structure Ref_Exc_Case_Thms = Named_Thms
    (val name = @{binding "ref_exc_case"}
     val description = "Theorems for automatic refinement over exceptions_case")
                               
structure Ref_If_Thms = Named_Thms
    (val name = @{binding "ref_if"}
     val description = "Theorems for automatic refinement over if-then-else")

structure Ref_If_Full_Thms = Named_Thms
    (val name = @{binding "ref_if_full"}
     val description = "Theorems for automatic refinement over if-then-else, with different conditions")

structure Ref_Choice_Thms = Named_Thms
    (val name = @{binding "ref_choice"}
     val description = "Theorems for automatic refinement over choice")

structure Ref_MapM_Thms = Named_Thms
    (val name = @{binding "ref_mapM"}
     val description = "Theorems for automatic refinement over mapM")

structure Ref_Bind_BindS_Thms = Named_Thms
    (val name = @{binding "ref_bind_bindS"}
     val description = "Theorems for automatic refinement over bind/bindS")

structure Ref_Try_Catch_BindS_Thms = Named_Thms
    (val name = @{binding "ref_try_catch_bindS"}
     val description = "Theorems for automatic refinement over try_catch/bindS")

structure Ref_MapM_BindS_Thms = Named_Thms
    (val name = @{binding "ref_mapM_bindS"}
     val description = "Theorems for automatic refinement over mapM/bindS")

structure Ref_BindS_Thms = Named_Thms
    (val name = @{binding "ref_bindS"}
     val description = "Theorems for automatic refinement over bindS")

structure Ref_Sim_Defs = Named_Thms
    (val name = @{binding "ref_sim_def"}
     val description = "Simulation relation definitions for simplification during automatic refinement")

structure Ref_Simp_Thms = Named_Thms
    (val name = @{binding "ref_simp"}
     val description = "Theorems for simplification during automatic refinement")

structure Ref_Cong_Thms = Named_Thms
    (val name = @{binding "ref_cong"}
     val description = "Congruence theorems for simplification during automatic refinement")

val get_bind_thms = Ref_Bind_Thms.get
val get_try_catch_thms = Ref_Try_Catch_Thms.get
val get_exc_case_thms = Ref_Exc_Case_Thms.get
val get_if_thms = Ref_If_Thms.get
val get_if_full_thms = Ref_If_Full_Thms.get
val get_choice_thms = Ref_Choice_Thms.get
val get_mapM_thms = Ref_MapM_Thms.get
val get_bind_bindS_thms = Ref_Bind_BindS_Thms.get
val get_try_catch_bindS_thms = Ref_Try_Catch_BindS_Thms.get
val get_mapM_bindS_thms = Ref_MapM_BindS_Thms.get
val get_bindS_thms = Ref_BindS_Thms.get
val get_sim_defs = Ref_Sim_Defs.get
val get_simp_thms = Ref_Simp_Thms.get
val get_cong_thms = Ref_Cong_Thms.get

val _ = (Context.>> o Context.map_theory) Ref_Bind_Thms.setup
val _ = (Context.>> o Context.map_theory) Ref_Try_Catch_Thms.setup
val _ = (Context.>> o Context.map_theory) Ref_Exc_Case_Thms.setup
val _ = (Context.>> o Context.map_theory) Ref_If_Thms.setup
val _ = (Context.>> o Context.map_theory) Ref_If_Full_Thms.setup
val _ = (Context.>> o Context.map_theory) Ref_Choice_Thms.setup
val _ = (Context.>> o Context.map_theory) Ref_MapM_Thms.setup
val _ = (Context.>> o Context.map_theory) Ref_Bind_BindS_Thms.setup
val _ = (Context.>> o Context.map_theory) Ref_Try_Catch_BindS_Thms.setup
val _ = (Context.>> o Context.map_theory) Ref_MapM_BindS_Thms.setup
val _ = (Context.>> o Context.map_theory) Ref_BindS_Thms.setup
val _ = (Context.>> o Context.map_theory) Ref_Sim_Defs.setup
val _ = (Context.>> o Context.map_theory) Ref_Simp_Thms.setup
val _ = (Context.>> o Context.map_theory) Ref_Cong_Thms.setup

fun pair_to_list (a, b) = [a, b]

val eq_term_pair = eq_pair (op aconv) (op aconv)

(* Storing terms with Abs outside leads to many spurious matches when
 looking refinement theorems up; we thus strip this layer when inserting
 and looking up terms
*)
val remove_abs = snd o mk_abs_var 

structure RefThms = Generic_Data
(
  type T = (term * thm list) Item_Net.T;
  val empty : T = Item_Net.init (Term.could_unify o pairself fst) (single o remove_abs o fst);
  val extend = I;
  fun merge tabs : T = Item_Net.merge tabs;
)

fun ap_fn_pair (f, g) x = (f x, g x)

fun extract_sim_targets t = let
  val head = head_of t

  fun extract_term (Const (@{const_name "sim"}, _)) = (get_arg 1, get_arg 0)
    | extract_term (Const (@{const_name "sim_norm"}, _)) = (get_arg 1, get_arg 0) 
    | extract_term (Const (@{const_name "sim_normN"}, _)) = (get_arg 1, get_arg 0) 
    | extract_term (Const (@{const_name "sim_normE"}, _)) = (get_arg 1, get_arg 0) 
    | extract_term (Const (@{const_name "sim_ext"}, _)) = (get_arg 4, get_arg 3) 
    | extract_term (Const (@{const_name "sim_norm_ext"}, _)) = (get_arg 3, get_arg 2) 
    | extract_term t = raise TERM ("Unknown simulation type", [t])

in
  ap_fn_pair (extract_term head) t
end

val extract_sim_targets_thm =
    concl_of
        #> HOLogic.dest_Trueprop
        #> extract_sim_targets
 
fun thm_to_item thm = (extract_sim_targets_thm thm |> snd |> mk_abs_var |> snd, thm)

val all_thms = flat o map snd o Item_Net.content o RefThms.get;

val get = all_thms o Context.Proof;

val get_full = Item_Net.content o RefThms.get o Context.Proof;

val transfer = RefThms.put o RefThms.get o Context.Proof

fun get_term_ref' context t =
  (
    remove_abs t
    |> Item_Net.retrieve (RefThms.get context) 

  (* TODO: we now flatten all the results, but it's not clear what
    causes multiple results to appear in the first place. See
    cl-isabelle-users post on get_info for a similar issue *)
    (*
    |> tap (fn results =>
               if length results > 1
               then (
                   "ref_thms: Warning: " ^ PolyML.makestring (length results) ^ " results for"
                                               |> tracing;
                   tracing (PolyML.makestring (pairself head_of ts))
                   )
               else ()
           )
     *)
    |> map snd
    |> flat
    |> distinct Thm.eq_thm
  ) handle Empty => [];

val get_term_ref = get_term_ref' o Context.Proof

fun get_old_thms thm context = let
    val (t, _) = thm_to_item thm
in
    (t, get_term_ref' context t)
end
    
fun add_thm thm context = let
    val (t, old_thms) = get_old_thms thm context
    val new_thms = Thm.add_thm thm old_thms
    val update = Item_Net.update (t, new_thms)
in
    RefThms.map update context
end

fun add_thms [] context = context
  | add_thms (thms as (thm::_)) context = let
        val (t, old_thms) = get_old_thms thm context
        val new_thms = fold Thm.add_thm thms old_thms
        val update = Item_Net.update (t, new_thms)
    in
        RefThms.map update context
    end

fun del_thm thm context = let
    val (t, old_thms) = get_old_thms thm context
    val new_thms = Thm.del_thm thm old_thms
    val update = Item_Net.update (t, new_thms)
in
    RefThms.map update context
end

val add = Thm.declaration_attribute add_thm;
val del = Thm.declaration_attribute del_thm;

end

