import Control.Monad.Error (ErrorT(ErrorT), runErrorT, Error, noMsg)
import Control.Monad.State (StateT(StateT), runStateT)
import Control.Monad.Trans.Maybe (MaybeT(MaybeT), runMaybeT)
import Data.Set (Set, insert, member, union, empty, singleton, deleteFindMin, fromList, toList, isSubsetOf)
import qualified Data.Set as SetM
import Data.Maybe (isJust, isNothing, fromMaybe)
import Data.List (nub)
import Control.Monad (filterM)


data Exception = FailWrite | FailLLRead | FailRead | FailMode | Degraded | DegradedH | DefaultErr
    deriving (Eq, Ord, Show)

instance Error Exception where
    noMsg = DefaultErr

data Content = Ok1 | Ok2 | Corrupt
    deriving (Eq, Ord, Show)

data Region = Primary | Secondary
    deriving (Eq, Ord, Show)

data State = S {
    primary :: Content,
    secondary :: Content,
    dead :: Set Region,
    unwritable :: Set Region
} deriving (Eq, Ord, Show)

powerset s
    | s == empty = singleton empty
    | otherwise = SetM.map (insert x) pxs `union` pxs
        where (x, xs) = deleteFindMin s
              pxs = powerset xs

powerList :: [a] -> [[a]]
powerList = filterM (const [False, True])

printList :: Show a => [a] -> IO ()
printList = putStrLn . unlines . map show

cnts = [Corrupt, Ok1, Ok2]
regs = [Primary, Secondary]

all_states = nub $ do
    p <- [Ok1, Corrupt]
    s <- (if p == Corrupt then [Corrupt, Ok1] else cnts)
    d <- powerList regs
    other <- powerList regs
    let unw = nub (d ++ other)
    return (S p s (fromList d) (fromList unw))


type RMonad sv = (ErrorT Exception (MaybeT (StateT (sv, State) [])))

mkRMonad = ErrorT . MaybeT . StateT
runRMonad = runStateT . runMaybeT . runErrorT

m1 <+> m2 = mkRMonad (\s ->
    runRMonad m1 s ++ runRMonad m2 s)

restart = mkRMonad $ \(sv, sp) -> [(Nothing, (sv, sp))]

throwE e = mkRMonad $ \(sv, sp) -> [(Just (Left e), (sv, sp))]

maybeIdN (Nothing, x) = (Nothing, x)

projectR :: RMonad sv a -> RMonad sv b
projectR m = mkRMonad $ \s -> 
    map maybeIdN $ filter (isNothing . fst) (runRMonad m s)

projectN :: RMonad sv a -> RMonad sv a
projectN m = mkRMonad $ \s -> 
    filter (isJust . fst) (runRMonad m s)

isLeft (Left _)     = True
isLeft (Right _)    = False

isRight (Right _)   = True
isRight (Left  _)   = False

fromLeft (Left x) = x
fromRight (Right x) = x

projectNN m = mkRMonad $ \s ->
    filter (isRight . fromMaybe (Left undefined) . fst) (runRMonad m s)

projectNE m = mkRMonad $ \s ->
    filter (isLeft . fromMaybe (Right undefined) . fst) (runRMonad m s)

--groupStartStates :: (Ord a, Ord sv) => sv -> RMonad sv a -> IO ()
groupStartResStates sv m = do
    let groups = groupSameRes sv m 
    let printRead = map formatSRGroup groups
    putStrLn (unlines printRead)

groupStartStates sv m = do
    let groups = map fst $ groupSameRes sv m 
    let printRead = map (formatGroup "--------------\n") groups
    putStrLn (unlines printRead)

formatGroup s l = s ++ unlines (map show l) ++ "\n"

formatSRGroup (starts, ends) = sLines ++ eLines
    where
        sLines = formatGroup "-----------------------\n" starts
        eLines = formatGroup "R-----------------------\n" ends

groupBySecond l = (map fst l, snd $ head l)

groupSameRes :: (Ord a, Ord sv) => sv -> RMonad sv a -> 
    [([State], [(Maybe (Either Exception a), (sv, State))])]
groupSameRes sv m = map (\(x, y) -> (x, toList y)) $ map groupBySecond $ groupByGlobal sameSets resultSets
    where
        resultSets = map (\sp -> (sp, fromList $ runRMonad m (sv, sp))) all_states
        sameSets (_, rs1) (_, rs2) = SetM.map normalize rs1 == SetM.map normalize rs2
        normalize (a, (sv, (S Ok2 Ok1 d uw))) = (a, (sv, S Ok1 Ok2 d uw))
        normalize x = x

groupByGlobal :: (a -> a -> Bool) -> [a] -> [[a]]
groupByGlobal p [] = []
groupByGlobal p (x:xs) = (x : matching) : groupByGlobal p nonMatching
    where
        matching = filter (p x) xs
        nonMatching = filter (not . p x) xs


catchR :: RMonad sv a -> RMonad sv a -> RMonad sv a
m `catchR` h = mkRMonad $ \s ->
    let 
        normal = runRMonad (projectN m) s
        restarting = runRMonad (projectR m) s
        caught = concatMap (runRMonad h . snd) restarting
    in
    normal ++ caught

catchE :: RMonad sv a -> (Exception -> RMonad sv a) -> RMonad sv a
m `catchE` h = mkRMonad $ \s ->
    let 
        normal = runRMonad (projectNN m) s
        restarting = runRMonad (projectR m) s
        exceptional = runRMonad (projectNE m) s
        caught = concatMap (\(Just (Left e), s) -> runRMonad (h e) s) exceptional
    in
        normal ++ restarting ++ caught

write_success :: Region -> Content -> RMonad sv ()
write_success inst cnt = mkRMonad $ \(sv, sp) ->
    if inst `member` (unwritable sp) 
        then []
        else if (inst == Primary) 
            then [(Just (Right ()), (sv, sp { primary = cnt }))]
            else [(Just (Right ()), (sv, sp { secondary = cnt }))]

write_fail_instance :: Region -> Content -> RMonad sv a
write_fail_instance inst cnt = mkRMonad $ \(sv, sp) ->
    if inst == Primary
        then [(Just (Left FailWrite), 
            (sv, sp { primary = Corrupt, unwritable = insert Primary (unwritable sp) }))]
        else [(Just (Left FailWrite), 
            (sv, sp { secondary = Corrupt, unwritable = insert Secondary (unwritable sp) }))]

write_corrupt_restart :: Region -> Content -> RMonad sv a
write_corrupt_restart inst cnt = mkRMonad $ \(sv, sp) ->
    if inst == Primary 
        then [(Nothing, (sv, sp { primary = Corrupt }))]
        else [(Nothing, (sv, sp { secondary = Corrupt }))]

write_fail_readonly :: Region -> Content -> RMonad sv a
write_fail_readonly inst cnt = mkRMonad $ \(sv, sp) ->
    [(Just (Left FailWrite), (sv, sp { unwritable = insert inst (unwritable sp) }))]

write_f inst cnt = do {
    write_success inst cnt 
    <+> write_fail_instance inst cnt
    <+> write_corrupt_restart inst cnt
    <+> write_fail_readonly inst cnt
    <+> restart `catchE` (\e -> throwE e <+> restart);
    return () <+> restart
    }   

ll_read_success :: Region -> RMonad sv Content
ll_read_success inst = mkRMonad $ \(sv, sp) ->
    if inst `member` (dead sp)
        then []
        else if inst == Primary
            then [(Just (Right (primary sp)), (sv, sp))]
            else [(Just (Right (secondary sp)), (sv, sp))]

ll_read_failure :: Region -> RMonad sv Content
ll_read_failure inst = mkRMonad $ \(sv, sp) ->
    [(Just (Left FailLLRead), 
        (sv, sp { dead = insert inst (dead sp)}))]
    
ll_read_corrupt :: Region -> RMonad sv Content
ll_read_corrupt inst = mkRMonad $ \(sv, sp) ->
    if inst `member` (dead sp)
        then []
        else if inst == Primary
            then [(Just (Right (Corrupt)), 
                (sv, sp { primary = Corrupt }))]
            else [(Just (Right (Corrupt)), 
                (sv, sp { secondary = Corrupt }))]
    
ll_read inst = do {
    x <- ((ll_read_success inst
    <+> ll_read_failure inst
    <+> ll_read_corrupt inst
    <+> restart) `catchE` (\e -> throwE e <+> restart));
    return x <+> restart
    }
    

ok_hash cnt = cnt /= Corrupt

get_raw_hashes inst = do
    cnt <- ll_read inst
    if ok_hash cnt
        then return cnt
        else throwE FailRead
    
get_hash_table = get_raw_hashes

write_hash_table inst cnt = write_f inst cnt

get_hashes = get_hash_table Secondary
   `catchE` (\e -> 
        case e of
            FailLLRead -> throwE DegradedH
            FailRead -> (
                (do
                    ht <- get_hash_table Primary
                    write_hash_table Secondary ht
                    return ht
                ) 
                `catchE` (\e ->
                    case e of
                        FailLLRead -> throwE FailMode
                        _ -> throwE e)
                )
            _ -> throwE e
    )    

write_hashes cnt = (do
        write_f Secondary cnt
        write_f Primary cnt
    ) `catchE` (\e -> case e of
        FailWrite -> throwE DegradedH
        _ -> throwE e
    )

sync_hashes :: Region -> Region -> RMonad sv ()
sync_hashes to from = do
    h_cnt <- (get_raw_hashes from `catchE` const (throwE FailMode))
    write_f to h_cnt `catchE` const (throwE DegradedH)

check_hash_region = (do
        h_cnt2 <- get_raw_hashes Secondary
        (do
            h_cnt1 <- get_raw_hashes Primary
            if h_cnt1 /= h_cnt2
                then sync_hashes Primary Secondary
                else return ()
         ) `catchE` (\e -> case e of
            FailRead -> sync_hashes Primary Secondary
            FailLLRead -> throwE DegradedH
            _ -> throwE e)
    ) `catchE` (\e -> case e of
        FailRead -> sync_hashes Secondary Primary
        FailLLRead -> do { 
            get_raw_hashes Primary `catchE` const (throwE FailMode);
            throwE DegradedH
        }
        _ -> throwE e)


maybe_degrade_primary = 
    write_fail_readonly Primary Ok1 `catchE` const (return ()) 
    <+> return ()

check2 = runRMonad $ (projectR check_hash_region) `catchR` (projectNN check_hash_region)
-- check1 = runRMonad $ projectNN check_hash_region
check1 = runRMonad $ do {projectNN check_hash_region; maybe_degrade_primary}

reduction1 = zipWith (==) (map (fromList. check1) whole_states) (map (fromList . check2) whole_states)
    where 
        whole_states = zip (repeat 1) (all_states)

-- Relax condition: test inclusion instead of equality
reduction_sub = zipWith (isSubsetOf) (map (fromList . check2) whole_states) (map (fromList. check1) whole_states)
    where 
        whole_states = zip (repeat 1) (all_states)


{-
 - Potential problem:
 - 1. Hashes are correct but differ in the initial state
 - 2. We try to sync Sec -> Prim
 - 3. Prim fails as unwritable
 - 4. We restart, but now Sec fails
 - 5. We end up with both synced, but Prim being (silently) unwritable
-}
problem = all_states !! 18
